import numpy as np
import numpy.polynomial.polynomial as npol
import scipy.linalg as la
import scipy.optimize as op

import copy
import itertools
import types


class monicPolynomial(npol.Polynomial):
    def __init__(self, coef, coef_type='monic'):
        if len(coef) == 0:
            coef = [1.]
        elif coef_type == 'monic':
            if type(coef) == list:
                coef = coef + [1.]
            else:
                coef = np.concatenate((coef, np.array([1.])))
        elif coef_type == 'zeros':
            coef = npol.polyfromroots(coef)
        elif coef_type == 'normal':
            if np.abs(coef[-1] - 1.) > 1e-15:
                raise Exception('Coefficients don\'t define a monic polynomial')
        else:
            raise Exception('Invalid coef_type, should be \'normal\', \'monic\' or \'zeros\'.')
        super().__init__(coef)

    def f_polynomial(self):
        if len(self.coef) == 1:
            # zero'th degree polynomial is 1
            def p(x):
                if type(x) == np.ndarray:
                    return np.ones(x.shape, dtype=complex)
                else:
                    return 1.+0.*1j
        elif len(self.coef) == 2:
            # first degree polynomial, only one coefficient
            p = lambda x: x + self.coef[0]
        else:
            # higher degree polynomial
            powers = np.arange(len(self.coef))
            def p(x):
                if type(x) == np.ndarray:
                    return np.sum(self.coef[np.newaxis,:] * x[:,np.newaxis]**powers[np.newaxis,:], 1)
                else:
                    return np.sum(self.coef * x**powers)
        return p


def pf_is_left_vec(p1, p2, point):
    """
    Tests if a point is Left|On|Right of the i'th trough the points 
    in p1[i,:] and p2[i,:], for all i.

    Adapted from Maciej Kalisiak <mac@dgp.toronto.edu>, the January 
    2001 Algorithm "Area of 2D and 3D Triangles and Polygons"

    Input: 
        p1, p2: 2d numpy arrays of dimension v*2, where v is the number of
            vertices
        point: 1d numpy array of length 2, representing the point in the plane

    Return: numpy array of size v, its i'th valuas is
            >0 for point left of the line through p1[i,:] and p2[i,:]
            =0 for point on the line
            <0 for point right of the line
    """
    return (p2[:,0]-p1[:,0]) * (point[1]-p1[:,1]) - (point[0]-p1[:,0]) * (p2[:,1]-p1[:,1])


def pf_winding_number(point, poly):
    """
    Winding number test for a point in a polygon.
    (Franklin, 2000), adapted from Maciej Kalisiak <mac@dgp.toronto.edu>

    Input:  
        point: 1d numpy array of length two, representing a point in
            the plane
        poly: 2d numpy array of dimension v*2, with v the number of
            vertices
    
    Output: 
        wn: int, the winding number (=0 only if point is outside the polygon)
    """
    wn = 0
    # up intersections
    ind = np.where( np.logical_and( poly[:-1,1] <= point[1], poly[1:,1] > point[1] ) )[0]
    wn += np.where( pf_is_left_vec(poly[ind,:], poly[ind+1,:], point) > 0 )[0].shape[0]
    # down intersections
    ind = np.where( np.logical_and( poly[:-1,1] > point[1], poly[1:,1] <= point[1] ) )[0]
    wn -= np.where( pf_is_left_vec(poly[ind,:], poly[ind+1,:], point) <= 0 )[0].shape[0]

    return wn


class contour(object):
    def __init__(self, N_eval=1e3):
        self.curves =  [lambda x:               np.exp(2.*np.pi*1j*x)]
        self.dcurves = [lambda x: 2.*np.pi*1j * np.exp(2.*np.pi*1j*x)]
        self.N_eval = [N_eval]
        self.make_arrays = False

    def construct_arrays(self):
        self.ts = [np.linspace(0., 1., int(N)) for N in self.N_eval]
        self.curve_arrs = [curve(self.ts[i]) for i, curve in enumerate(self.curves)]
        self.dcurve_arrs = [dcurve(self.ts[i]) for i, dcurve in enumerate(self.dcurves)]
        self.make_arrays = True

    def store_fun_vals(self, fun):
        self.fun_arrs  = [fun(curve_arr) for curve_arr in self.curve_arrs]

    def calc_boundaries(self):
        # boundaries
        minreal = 0.; maxreal = 0.
        minimag = 0.; maximag = 0.
        for i, N in enumerate(self.N_eval):
            if self.make_arrays:
                c = self.curve_arrs[i]
            else:
                c = self.curves[i](np.linspace(0., 1., int(N)))
            mincr = np.min(c.real); maxcr = np.max(c.real)
            if mincr < minreal: minreal = mincr
            if maxcr > maxreal: maxreal = maxcr
            minci = np.min(c.imag); maxci = np.max(c.imag)
            if minci < minimag: minimag = minci
            if maxci > maximag: maximag = maxci
        self.minreal = minreal; self.maxreal = maxreal
        self.minimag = minimag; self.maximag = maximag

    def construct_polygon(self):
        """ construct polygon from curves """
        if 'polygon' not in self.__dict__:
            if self.make_arrays:
                polygon = np.concatenate(self.curve_arrs)
            else:
                polygon = np.concatenate([curve(np.linspace(0., 1., int(self.N_eval[i]))) for i, curve in enumerate(self.curves)])
            self.polygon = np.concatenate((polygon.real[:,np.newaxis], polygon.imag[:,np.newaxis]), 1)
            self.polygon[-1,:] = copy.copy(self.polygon[0,:])

    def is_inside(self, cpoint):
        self.construct_polygon()
        return pf_winding_number(np.array([cpoint.real, cpoint.imag]), self.polygon) != 0


class circularContour(contour):
    def __init__(self, radius=1., center=0.+0j, N_eval=1e3):
        super().__init__()
        self.center = center
        self.radius = radius
        self.curves  = [lambda x: center +               radius * np.exp(2.*np.pi*1j*x)]
        self.dcurves = [lambda x:          2.*np.pi*1j * radius * np.exp(2.*np.pi*1j*x)]
        self.N_eval = [N_eval]

    def is_inside(self, cpoint):
        return np.abs(cpoint - self.center) < self.radius


class trapezoidContour(contour):
    def __init__(self, t1, nparam=1e3, N_eval=1e3):
        super().__init__()
        # set of points defining the trapezoid
        self.points = [t1[1],t1[0]] + [np.conj(t1[0]), np.conj(t1[1])]
        # line segments defining the trapezoid contour
        self.curves  = [lambda x, p1=p, p2=self.points[(i+1)%len(self.points)]: p1 + x * (p2-p1) for i, p in enumerate(self.points)]
        self.dcurves = [lambda x, p1=p, p2=self.points[(i+1)%len(self.points)]:          (p2-p1) for i, p in enumerate(self.points)]
        self.N_eval = [N_eval, N_eval, N_eval, N_eval]

    def divide_real_axis(self):
        ps = self.points
        prcut = (ps[0].real + ps[1].real) / 2.
        pccut = ps[0].imag + (ps[1].imag-ps[0].imag) * (prcut - ps[0].real) / (ps[1].real-ps[0].real)

        return trapezoidContour(t1=(ps[0], prcut+1j*pccut)), trapezoidContour(t1=(prcut+1j*pccut, ps[1]))


class poleFinder:
    def __init__(self, fun=lambda x:x, dfun=lambda x:1, 
                        global_poles={}, make_arrays=False, use_known_zeros=False, **kwargs):
        # callables: the function and its derivative
        self.fun = fun
        self.dfun = dfun
        if len(global_poles) > 0:
            self.has_global_poles = True
            self.global_poles = np.array(global_poles['poles'])
            self.global_pmultiplicities = np.array(global_poles['pmultiplicities'])
        else:
            self.has_global_poles = False
        self.global_zeros = np.array([])
        self.global_zmultiplicities = np.array([])
        self.inner_contours = []; self.secondary_contours = []
        # set curves if args are given
        if len(kwargs) > 0:
            if 'poles' in list(kwargs.keys()):
                if 'pmultiplicities' in list(kwargs.keys()):
                    self.set_contour(kwargs['contour'], make_arrays=make_arrays, use_known_zeros=use_known_zeros, 
                                        poles=kwargs['poles'], pmultiplicities=kwargs['pmultiplicities'])
                else:
                    self.set_contour(kwargs['contour'], make_arrays=make_arrays, use_known_zeros=use_known_zeros,
                                        poles=kwargs['poles'])

            else:
                self.set_contour(kwargs['contour'], make_arrays=make_arrays, use_known_zeros=use_known_zeros)

    def set_contour(self, contour, make_arrays=False, use_known_zeros=False,
                        internal_contours=[],
                        poles=[], pmultiplicities=[], zeros=[], zmultiplicities=[]):
        self.make_arrays = make_arrays
        self.use_known_zeros = use_known_zeros
        # callables
        self.contour = contour
        # arrays
        if self.make_arrays:
            self.contour.construct_arrays()
            # function arrays
            self.contour.store_fun_vals(lambda x: self.dfun(x)/self.fun(x))
            # self.fun_arrs  = [ self.fun(curve_arr) for curve_arr in self.contour.curve_arrs]
            # self.dfun_arrs = [self.dfun(curve_arr) for curve_arr in self.contour.curve_arrs]
        self.contour.calc_boundaries()
        # set poles within contour and their multiplicities
        if len(poles) == 0 and self.has_global_poles:
            # add poles that are inside curve
            self.poles = np.array([(i, pole) for i, pole in enumerate(self.global_poles) if self.contour.is_inside(pole)])
            if len(self.poles) > 0:
                self.pmultiplicities = self.global_pmultiplicities[self.poles[:,0].real.astype(int)]
                self.poles = self.poles[:,1]
            else:
                self.pmultiplicities = np.array([])
                self.poles = np.array([])
        elif len(poles) > 0:
            self.poles = np.array(poles)
            if len(pmultiplicities) == 0:
                self.pmultiplicities = np.ones(len(poles))
            else:
                self.pmultiplicities = np.array(pmultiplicities)
        else:
            self.poles = []
            self.pmultiplicities = []
        # calculate the known zeros that are within the contour
        if use_known_zeros and len(self.global_zeros) > 0:
            # add zeros that are inside curve
            self.zeros = np.array([(i, zero) for i, zero in enumerate(self.global_zeros) if self.contour.is_inside(zero)])
            if len(self.zeros) > 0:
                self.zmultiplicities = self.global_zmultiplicities[self.zeros[:,0].real.astype(int)]
                self.zeros = self.zeros[:,1]
            else:
                self.zmultiplicities = np.array([])
                self.zeros = np.array([])
        else:
            self.zmultiplicities = np.array([])
            self.zeros = np.array([])

    def add_secondary_contour(self, contour):
        """
        assumed to entirely inside or entirely outside the main contour, 
        and non-overlapping with the other inner contours
        """
        if self.make_arrays:
            contour.construct_arrays()
            contour.store_fun_vals(lambda x: self.dfun(x)/self.fun(x))
        self.secondary_contours.append(contour)
        self.inner_contours.append(contour)

    def inner_prod(self, p1, p2, compute_maxpsum=False):
        # construct the integrand either as a callable or an array
        if self.make_arrays:
            integrand = [p1(curve_arr)*p2(curve_arr)*self.contour.fun_arrs[i] for i, curve_arr in enumerate(self.contour.curve_arrs)]
        else:
            integrand = lambda s: p1(s)*p2(s)*self.dfun(s)/self.fun(s)
        # construct the poles
        if len(self.poles) > 0:
            sum_poles = np.sum(self.pmultiplicities * p1(self.poles) * p2(self.poles))
        else:
            sum_poles = 0.
        # construct the known zeros
        if self.use_known_zeros and len(self.global_zeros) > 0:
            sum_zeros = - np.sum(self.zmultiplicities * p1(self.zeros) * p2(self.zeros))
            # sum_zeros = 0
        else:
            sum_zeros = 0.
        if len(self.inner_contours) > 0:
            inner_sum = 0.
            for inner_contour in self.inner_contours:
                if self.make_arrays:
                    inner_integrand = [p1(curve_arr)*p2(curve_arr)*inner_contour.fun_arrs[i] for i, curve_arr in enumerate(inner_contour.curve_arrs)]
                else:
                    inner_integrand = lambda s: p1(s)*p2(s)*self.dfun(s)/self.fun(s)
                inner_sum -= self.contour_integral(fun=inner_integrand, contour=inner_contour)
        else:
            inner_sum = 0.
        # compute the contour integral
        if compute_maxpsum:
            a, b = self.contour_integral(fun=integrand, compute_maxpsum=compute_maxpsum)
            return a + sum_poles + sum_zeros + inner_sum, b
        else:
            return self.contour_integral(fun=integrand) + sum_poles + sum_zeros + inner_sum

    def generalized_hankel_matrices(self, phis, pols):
        G  = np.zeros((len(phis), len(phis)), dtype=complex)
        G1 = np.zeros((len(phis), len(phis)), dtype=complex)
        prow = [monicPolynomial(npol.polymul(pols[1].coef, pols[j].coef), coef_type='normal').f_polynomial() for j in range(len(phis))]
        for i,j in itertools.product(list(range(len(phis))), list(range(len(phis)))):
            G [i,j] = self.inner_prod(phis[i], phis[j])
            G1[i,j] = self.inner_prod(phis[i], prow[j])
        return G, G1

    def contour_integral(self, fun=None, contour=None, compute_maxpsum=False):
        """
        Computes the contour integral over the contour formed by self.curves

        Input:
            - fun: callable or list of 1d arrays, the latter case corresponding to the function
                evaluated over the curves in self.curves. If None, self.fun is chosen as function.
            - contour: the contour over which to integrate the function. If None, the self.contour
                is chosen
            - compute_maxpsum: boolean, whether to compute the maximal partial sum

        Output:
            - psum: complex, the integral value
            - maxpsum: float, the maximum partial sum
        """
        if self.make_arrays:
            if contour == None:
                contour = self.contour
            if isinstance(fun, types.FunctionType):
                fun = [fun(curve_arr) for curve_arr in contour.curve_arrs]
            elif fun == None:
                fun = self.contour.fun_arrs
            psum = 0.;
            if compute_maxpsum: maxpsum = 0.
            for i, t in enumerate(contour.ts):
                fval = fun[i]*contour.dcurve_arrs[i]
                dt = t[1]-t[0]
                if compute_maxpsum:
                    cumpsum = psum + dt * np.cumsum(fval[:-1] + fval[1:]) / (4.*np.pi*1j)
                    psum = cumpsum[-1]
                    nmaxpsum = np.max(np.abs(cumpsum))
                    if nmaxpsum > maxpsum: maxpsum = nmaxpsum
                else:
                    psum += dt * np.sum(fval[:-1] + fval[1:]) / (4.*np.pi*1j)
        else:
            if fun == None:
                fun = lambda x: self.dfun(x) / self.fun(x)
            if contour == None:
                contour = self.contour
            psum = 0.
            if compute_maxpsum: maxpsum = 0.
            for i, curve in enumerate(contour.curves):
                dcurve = contour.dcurves[i]
                dt = 1./contour.N_eval[i]
                if isinstance(fun, types.FunctionType):
                    t = np.linspace(0., 1., int(contour.N_eval[i]))
                    fval = fun(curve(t)) * dcurve(t)
                else:
                    fval = fun[i] * dcurve(t)
                if compute_maxpsum:
                    cumpsum = psum + dt * np.cumsum(fval[:-1] + fval[1:]) / (4.*np.pi*1j)
                    psum = cumpsum[-1]
                    nmaxpsum = np.max(np.abs(cumpsum))
                    if nmaxpsum > maxpsum: maxpsum = nmaxpsum
                else:
                    psum += dt * np.sum(fval[:-1] + fval[1:]) / (4.*np.pi*1j)
        if compute_maxpsum:
            return psum, maxpsum
        else:
            return psum

    def points_within_contour(self, points):
        """
        check if all points in points lie within the contour
        """
        for point in points:
            if point.real < self.contour.minreal or point.real > self.contour.maxreal:
                return False
            elif point.imag < self.contour.minimag or point.imag > self.contour.maximag:
                return False
            # else: 
            #     if self.make_arrays:
            #         p = np.array(self.curve_arrs).flatten()
            #     else:
            #         p = np.array([curve(self.ts[i]) for i, curve in enumerate(self.curves)]).flatten()
            #     polygon = np.concatenate((p.real[:,np.newaxis], p.imag[:,np.newaxis]), 1)
            #     if pf_winding_number(np.array([point.real, point.imag]), polygon) == 0:
            #         return False
        return True

    def test_contour(self, eps=1e-5, pprint=False):
        """
        Test if the contour integration on this contour is inaccurate (i.e. deviates
        from n for n in Z by more than eps). 
        """
        pol_unity = monicPolynomial([])
        p_unity = pol_unity.f_polynomial()
        residue = self.inner_prod(p_unity, p_unity)
        N = int(round(residue.real))
        accuracy = np.abs(N - residue)
        if pprint:
            print('residue =', residue)

        return accuracy < eps

    def find_zeros(self, pprint=False):
        """
        Find the zeros and poles of a complex function (C->C) inside a closed curve.

        input:
        """
        # total multiplicity of zeros (number of zeros * order)
        pol_unity = monicPolynomial([])
        p_unity = pol_unity.f_polynomial()
        residue = self.inner_prod(p_unity, p_unity)
        N = int(round(residue.real))
        accuracy = np.abs(N - residue)
        if pprint: print('N =', N, ' (accuracy =', accuracy, ')')
        # if N is zero
        if N == 0:
            n = 0
            zeros = np.array([])
        else:
            # list of FOPs
            phis = []; pols = []
            phis.append( p_unity ); pols.append( pol_unity ) # phi_0
            # compute arithmetic mean of nodes
            p_aux = monicPolynomial( [0.] ).f_polynomial()
            mu = self.inner_prod( p_unity, p_aux ) / N
            if pprint: print('mu =', mu)
            # append first polynomial
            pol = monicPolynomial( [-mu] )
            phis.append( pol.f_polynomial() ); pols.append( pol ) # phi_1
            # if the algorithm quits after the first zero, it is mu
            zeros = np.array([mu])
            n = 1
            # initialization
            r = 1; t = 0
            while r+t < N:
                if pprint: print(str(r+t))
                # naive criterion to check if phi_r+t+1 is regular
                prod_aux, maxpsum = self.inner_prod(phis[-1], phis[-1], compute_maxpsum=True)
                # compute eigenvalues of the next pencil
                G, G1 = self.generalized_hankel_matrices(phis, pols)
                eigv = la.eigvals(G1, G)
                # print eigv
                # check if these eigenvalues lie within the contour
                if self.points_within_contour(eigv+mu):
                # if np.abs(prod_aux)/maxpsum > eps_reg:
                    if pprint: print(str(r+t) + '.1')
                    # compute next FOP in the regular way
                    pol = monicPolynomial(eigv, coef_type='zeros')
                    phis.append(pol.f_polynomial()); pols.append(pol)
                    r += t+1; t = 0
                    n = r
                    zeros = eigv + mu
                else:
                    if pprint: print(str(r+t) + '.2')
                    c = npol.polyfromroots([mu])
                    pol = monicPolynomial(npol.polymul(c, pols[-1].coef), coef_type='normal')
                    pols.append(pol); phis.append(pol.f_polynomial())
                    t += 1

        # check multiplicities
        # constuct vandermonde system
        if len(zeros) > 0:
            A = np.vander(zeros).T[::-1,:]
            b = np.array([ self.inner_prod(p_unity, lambda x, k=k:x**k) for k in range(n)])
            # solve the Vandermonde system
            nu = la.solve(A, b)
            nu = np.round(nu).real.astype(int)
            sane = (np.sum(nu) == N)
        else:
            nu = np.array([])
            sane = True
        # for printing result
        if pprint:
            print('>>> Result of computation of zeros <<<')
            print('number of zeros = ', n - len(np.where(nu == 0)[0]))
            pstring = 'yes!' if sane == 1 else 'no!'
            print('sane? ' + pstring)
            for i, zero in enumerate(zeros):
                print('zero #' + str(i+1) + ' = ' + str(zero) + ' (multiplicity = ' + str(nu[i]) + ')')
            print('')
        # eliminate spurious zeros (the ones with zero multiplicity)
        inds_ = np.argsort(zeros.real)
        # inds = np.where(nu[inds_] > 0)[0]

        if self.use_known_zeros:
            self.global_zeros = np.concatenate((self.global_zeros, zeros[inds_][inds]))
            self.global_zmultiplicities = np.concatenate((self.global_zmultiplicities, nu[inds_][inds]))
            iglob = np.argsort(self.global_zeros.real)
            self.global_zeros = self.global_zeros[iglob]
            self.global_zmultiplicities = self.global_zmultiplicities[iglob]

        return zeros[inds_], (n-len(np.where(nu == 0)[0]), nu[inds_], sane)

    def find_real_zeros_recursively(self, realtol=1e-4, minradius=1., depth=0, maxdepth=10, contourparams_orig=None, pprint=False):
        """
        Searches for the zeros of a real function by using circular contours with a center on the real axis.
        If the starting contour is not accurate enough, it is split up in two equal parts.
        """
        # check if contour is circular and has real center
        assert isinstance(self.contour, circularContour)
        assert np.abs(self.contour.center.imag) < 1e-10*self.contour.radius
        # increase the depth parameter
        depth += 1
        if pprint: print('depth =', depth)
        # store the original contour parameters to restore this contour later on
        if contourparams_orig == None:
            contourparams_orig = (self.contour.center, self.contour.radius)
        # if the contour is sufficiently accurate, try to find the zeros within
        if self.test_contour(pprint=pprint):
            zeros, _ = self.find_zeros(pprint=pprint)
            realzero_inds = np.where(np.abs(zeros.imag) < realtol)[0]
            if len(realzero_inds) == 0 and len(zeros)%2 == 1:
                # if no real zeros are found, even though the number of zeros is uneven, look further
                zeros = self._continue_real_zero_recursion(depth=depth, maxdepth=maxdepth, contourparams_orig=contourparams_orig)
            else:
                zeros = zeros[realzero_inds].real
            # restore the original contour
            if contourparams_orig != None:
                contour = circularContour(center=contourparams_orig[0], radius=contourparams_orig[1])
                self.set_contour(contour, make_arrays=True)
            return zeros
        else:
            # look further for zeros
            zeros = self._continue_real_zero_recursion(depth=depth, maxdepth=maxdepth, contourparams_orig=contourparams_orig)
            # restore the original contour
            if contourparams_orig != None:
                contour = circularContour(center=contourparams_orig[0], radius=contourparams_orig[1])
                self.set_contour(contour, make_arrays=True)
            return zeros

    def _continue_real_zero_recursion(self, realtol=1e-4, minradius=1., depth=0, maxdepth=10, contourparams_orig=None, pprint=False):
        c0 = self.contour.center.real; r0 = self.contour.radius
        if r0 > minradius and depth < maxdepth:
            # first new contour for zero finding
            contour1 = circularContour(center=c0-r0/2.+0j, radius=r0/2.)
            self.set_contour(contour1, make_arrays=True)
            zeros1 = self.find_real_zeros_recursively(realtol=realtol, minradius=minradius, 
                                            depth=depth, maxdepth=maxdepth, contourparams_orig=contourparams_orig, pprint=pprint)
            # second new contour for zero finding
            contour2 = circularContour(center=c0+r0/2.+0j, radius=r0/2.)
            self.set_contour(contour2, make_arrays=True)
            zeros2 = self.find_real_zeros_recursively(realtol=realtol, minradius=minradius, 
                                            depth=depth, maxdepth=maxdepth, contourparams_orig=contourparams_orig, pprint=pprint)
            # zeros
            zeros = np.concatenate((zeros1, zeros2))
        else:
            zeros = np.array([])
        return zeros

    def find_zeros_and_poles_(self, eps_reg=1e-15, eps_stop=1e-10, P_estimate=0, pprint=False):
        """
        Find the zeros and poles of a complex function (C->C) inside a closed curve.

        input:
            -fun: callable, the function of which the poles have to be found
            -dfun: callable, the derivative of the function
            -contours: list of callables, the contours (R->C) that constitute
                a closed curve in the complex plane
            -dcontours: list of callables, the derivatives (R->C) of the
                respective contours
            -t_params: list of arrays, the parametrizations of the contours

        !!! Hard to get stopping right !!!
        """
        # total multiplicity of zeros (number of zeros * order)
        pol_unity = monicPolynomial([])
        p_unity = pol_unity.f_polynomial()
        s0 = int(round(self.inner_prod(p_unity, p_unity).real))
        N =  s0 + 2*P_estimate
        if pprint: print('N =', N)
        # list of FOPs
        phis = []; pols = []
        phis.append( p_unity ); pols.append( pol_unity ) # phi_0
        # if s0 is zero
        if s0 == 0:
            n = 0
            # arithmetic mean of nodes is zero
            mu = 0.
            pol = monicPolynomial( [0.] )
            phis.append( pol.f_polynomial() ); pols.append( pol ) # phi_1
            zeros = np.array([])
            r = 0; t = 1
        else:
            # compute arithmetic mean of nodes
            p_aux = monicPolynomial( [0.] ).f_polynomial()
            mu = self.inner_prod( p_unity, p_aux ) / N
            if pprint: print('mu =', mu)
            # append first polynomial
            pol = monicPolynomial( [-mu] )
            phis.append( pol.f_polynomial() ); pols.append( pol ) # phi_1
            # if the algorithm quits after the first zero, it is mu
            zeros = np.array([mu])
            n = 1
            # initialization
            r = 1; t = 0
        while r+t < N:
            if pprint: print('1.')
            # naive criterion to check if phi_r+t+1 is regular
            prod_aux, maxpsum = self.inner_prod( phis[-1], phis[-1], compute_maxpsum=True )
            # compute eigenvalues of the next pencil
            G, G1 = self.generalized_hankel_matrices(phis, pols)
            eigv = la.eigvals(G1, G)
            # check if these eigenvalues lie within the contour
            if self.points_within_contour(eigv+mu):
            # if np.abs(prod_aux)/maxpsum > eps_reg:
                if pprint: print('1.1')
                # compute next FOP in the regular way
                pol = monicPolynomial(eigv, coef_type='zeros')
                phis.append(pol.f_polynomial()); pols.append(pol)
                r += t+1; t = 0
                allsmall = True; tau = 0
                while allsmall and (r+tau) < N:
                    if pprint: print('1.1.1')
                    # if all further inner porducts are zero
                    taupol = npol.polyfromroots([mu for _ in range(tau)])
                    tauphi = monicPolynomial(npol.polymul(taupol, pols[-1].coef), coef_type='normal').f_polynomial()
                    ip, maxpsum = self.inner_prod(tauphi, phis[r], compute_maxpsum=True )
                    tau += 1
                    if np.abs(ip)/maxpsum > eps_stop:
                        allsmall = False
                if allsmall:
                    if pprint: print('1.1.2: STOP')
                    n = r
                    zeros = eigv + mu
                    t = N # STOP
            else:
                if pprint: print('1.2')
                c = npol.polyfromroots([mu])
                pol = monicPolynomial(npol.polymul(c, pols[-1].coef), coef_type='normal')
                pols.append(pol); phis.append(pol.f_polynomial())
                t += 1

        # check multiplicities
        # constuct vandermonde system
        A = np.vander(zeros).T[::-1,:]
        b = np.array([ self.inner_prod(p_unity, lambda x, k=k:x**k) for k in range(n)])
        # solve the Vandermonde system
        nu = la.solve(A, b)
        nu = np.round(nu).real.astype(int)
        # for printing result
        if pprint:
            print('>>> Result of computation of zeros <<<')
            print('number of zeros = ', n - len(np.where(nu == 0)[0]))
            for i, zero in enumerate(zeros):
                print('zero #' + str(i+1) + ' = ' + str(zero) + ' (multiplicity = ' + str(nu[i]) + ')')
            print('')
        # eliminate spurious zeros (the ones with zero multiplicity)
        inds = np.where(nu > 0)[0]
        return zeros[inds], n, nu[inds]

    def find_real_zeros(self, xtol=1e-9, vmin=None, vmax=None):
        if vmin == None:
            vmin = 0.
        if vmax == None:
            vmax = self.global_poles[-1]
        inds = np.where(np.logical_and(
                            np.logical_and(self.global_poles > vmin, self.global_poles < vmax),
                                        self.global_poles.imag < 1e-1) )[0]

        f_aux = lambda x: self.fun(x+0j).real
        df_aux = lambda x: self.dfun(x+0j).real

        # find the zeros
        zeros = []
        plist = np.concatenate(([vmin], self.global_poles[inds], [vmax]))
        for i in range(len(plist)-1):
            xeps = (plist[i+1]-plist[i])*xtol
            xmin = plist[i] + xeps; xmax = plist[i+1] - xeps
            # check if interval between poles is too small
            if xmin < xmax:
                # find zeros, maximal detection is 2 per interval
                sf0 = np.sign(f_aux(xmin)); sf1 = np.sign(f_aux(xmax))
                sdf0 = np.sign(df_aux(xmin)); sdf1 = np.sign(df_aux(xmax))
                xs = []
                if sf0 != sf1:
                    x0 = op.brentq(f_aux, xmin, xmax)
                    xs = [x0]
                elif sdf0 != sdf1:
                    dx0 = op.brentq(df_aux, xmin, xmax)
                    sfex = np.sign(f_aux(dx0))
                    if sfex != sf0:
                        x0 = op.brentq(f_aux, xmin, dx0)
                        # try:
                        x1 = op.brentq(f_aux, dx0, xmax)
                        # except ValueError as e:
                        #     print e
                        #     print self.fun(xmin), self.fun(xmax)
                        #     print np.sign(self.fun(xmin)), np.sign(self.fun(xmax))
                        #     print dx0
                        #     print self.fun(dx0)
                        #     print np.sign(self.fun(dx0))
                        xs = [x0, x1]
                zeros.extend(xs)
            # STILL TO BE IMPLEMTED: strategy to check if there are zeros
            # when interval is too small

        return np.sort(zeros), np.ones(len(zeros))


def find_zeros_on_segment(zeros, zmultiplicities, xmin, xmax, fun, dfun, poles, pmultiplicities, xtree='', pprint=False):
    """
    Auxiliary recursive function to find the zeros on a segment with sufficient accuracy. Decrease the contour radius 
    untill sufficient accuracy is reached
    """
    cc = circularContour(radius=(xmax-xmin)/2., center=(xmax+xmin)/2.+0j, N_eval=1e2)
    PF = poleFinder(fun=fun, dfun=dfun, global_poles={'poles': poles, 'pmultiplicities': pmultiplicities},
                            make_arrays=True, use_known_zeros=False, contour=cc)

    if pprint:
        # print ''
        # print '>>> <<<'
        print(xtree)
        print('xmin = ', xmin, ', xmax = ', xmax)
        print('radius = ', (xmax-xmin)/2., ', center = ', (xmax+xmin)/2.)
        print(PF.test_contour(pprint=pprint))

    if PF.test_contour():
        # find the zero
        p1, info = PF.find_zeros(pprint=False)
        zeros.extend(p1.tolist()); zmultiplicities.extend(info[1].tolist())
    else:
        xmin0 = xmin; xmax0 = (xmax+xmin)/2.
        xmin1 = xmax0; xmax1 = xmax

        find_zeros_on_segment(zeros, zmultiplicities, xmin0, xmax0, fun, dfun, poles, pmultiplicities, xtree=xtree+'0', pprint=pprint)
        find_zeros_on_segment(zeros, zmultiplicities, xmin1, xmax1, fun, dfun, poles, pmultiplicities, xtree=xtree+'1', pprint=pprint)

# make sure a variable is complex
def _to_complex(x):
    if isinstance(x, np.ndarray):
        if x.dtype == float:
            x = copy.copy(x)
            x = x + 0j
    elif type(x) == float:
        x = copy.copy(x)
        x = x + 0j
    return x



from neat.channels.ionchannels import IonChannel


class TestChannel(IonChannel):
    """
    Simple channel to test basic functionality
    """
    def define(self):
        # open probability
        self.p_open = '5 * a00**3 * a01**3 * a02 + a10**2 * a11**2 * a12'

        # state variables activation
        self.varinf = {}
        self.varinf['a00'] = '1. / (1. + exp((v-30.) / 100.))'
        self.varinf['a01'] = '1. / (1. + exp((-v+30.)/ 100.))'
        self.varinf['a02'] = '-10.'
        self.varinf['a10'] = '2. / (1. + exp((v-30.) / 100.))'
        self.varinf['a11'] = '2. / (1. + exp((-v+30.)/ 100.))'
        self.varinf['a12'] = '-30.'
        # state variable time scales
        self.tauinf = {}
        self.tauinf['a00'] = '1.'
        self.tauinf['a01'] = '2.'
        self.tauinf['a02'] = '1.'
        self.tauinf['a10'] = '2.'
        self.tauinf['a11'] = '2.'
        self.tauinf['a12'] = '3.'
        # default reversal
        self.e = -23.


class TestChannel2(IonChannel):
    """
    Simple channel to test basic functionality
    """
    def define(self):
        # open probability
        self.p_open = '.9 * a00**3 * a01**2 + .1 * a10**2 * a11'
        # state variables
        self.varinf = {'a00': '.3', 'a01': '.5',
                       'a10': '.4', 'a11': '.6'}
        self.tauinf = {'a00': '1.', 'a01': '2.',
                       'a10': '2.', 'a11': '2.'}
        # default reversal
        self.e = -23.


class Na_Ta(IonChannel):
    def define(self):
        """
        (Colbert and Pan, 2002)

        Used in (Hay, 2011)
        """
        self.ion = 'na'
        # concentrations the ion channel depends on
        self.conc = {}
        # define channel open probability
        self.p_open = 'h * m ** 3'
        # define activation functions
        self.alpha, self.beta = {}, {}
        self.alpha['m'] =  '0.182 * (v + 38.) / (1. - exp(-(v + 38.) / 6.))' # 1/ms
        self.beta['m']  = '-0.124 * (v + 38.) / (1. - exp( (v + 38.) / 6.))' # 1/ms
        self.alpha['h'] = '-0.015 * (v + 66.) / (1. - exp( (v + 66.) / 6.))' # 1/ms
        self.beta['h']  =  '0.015 * (v + 66.) / (1. - exp(-(v + 66.) / 6.))' # 1/ms
        # temperature factor for time-scale
        self.q10 = 2.95


class Kv3_1(IonChannel):
    def define(self):
        self.ion = 'k'
        # define channel open probability
        self.p_open = 'm'
        # define activation functions
        self.varinf = {'m': '1. / (1. + exp(-(v - 18.70) /  9.70))'}
        self.tauinf = {'m': '4. / (1. + exp(-(v + 46.56) / 44.14))'} # ms


class SK(IonChannel):
    def define(self):
        '''
        SK-type calcium-activated potassium current (Kohler et al., 1996)
        used in (Hay et al., 2011)
        '''
        self.ion = 'k'
        self.conc = ['ca']
        # define channel open probability
        self.p_open = 'z'
        # activation functions
        self.varinf = {'z': '1. / (1. + (0.00043 / ca)**4.8)'}
        self.tauinf = {'z': '1.'} # ms


class h(IonChannel):
    def define(self):
        """
        Hcn channel from (Bal and Oertel, 2000)
        """
        # define channel open probability
        self.p_open = '.8 * hf + .2 * hs'
        # define activation functions
        self.varinf, self.tauinf = {}, {}
        self.varinf['hf'] = '1. / (1. + exp((v + 82.) / 7.))'
        self.varinf['hs'] = '1. / (1. + exp((v + 82.) / 7.))'
        self.tauinf['hf'] = '40.'
        self.tauinf['hs'] = '300.'
        # default reversal
        self.e = -43.



import numpy as np
import scipy.linalg as la
import scipy.optimize as so

import copy


class MatCollection(object):
    def __init__(self, rmat, rvec, rvar, tvec, tvar):
        self.rmat = rmat
        self.rvec = rvec
        self.rvar = rvar
        self.tmat = tvec
        self.tvec = tvar

    def calcFVar(self):
        return self.rvar

    def calcJFRow(self):
        # print 't mat ='
        # print self.tmat
        # print 't vec ='
        # print self.tvec
        return self.tvec - np.dot(self.tmat, la.solve(self.rmat, self.rvec))
        # return - np.dot(self.tmat, la.solve(self.rmat, self.rvec))
        # return self.tvec - np.dot(self.tmat, np.dot(la.inv(self.rmat), self.rvec))


class NewtonMatrices(object):
    def __init__(self):
        self.initMatrixCollection()

    def initMatrixCollection(self):
        self.matrix_collections = []

    def addCollection(self, rmat, rvec, rvar, tvec, tvar):
        mc = MatCollection(rmat, rvec, rvar, tvec, tvar)
        self.matrix_collections.append(mc)

    def getFC(self):
        return np.array([mc.calcFVar() for mc in self.matrix_collections])

    def getJFC(self):
        return np.array([mc.calcJFRow() for mc in self.matrix_collections])

    def calcResidual(self):
        # rr = self.getFC()
        # print '!! r_nn =', rr
        return la.norm(self.getFC())


class IEPSolver(object):
    def __init__(self, pencil, lambdas=None):
        assert pencil.shape[1] == pencil.shape[2]

        self.pencil = pencil
        # shape constants
        self.K = self.pencil.shape[0]
        self.M = self.pencil.shape[2]
        # matrixcollection
        self.nm = NewtonMatrices()
        # target eigenvalues
        if lambdas is not None:
            self.initLambdas(lambdas)

    def initLambdas(self, lambdas):
        self.lambdas = lambdas
        self.N = self.lambdas.shape[0]

    def evalPencil(self, cvec):
        return self.pencil[0] + np.sum([c*p_mat for c, p_mat in zip(cvec, self.pencil[1:])], axis=0)

    def _evalQRs(self, cvec):
        NN, MM, KK = self.N, self.M, self.K

        self.nm.initMatrixCollection()
        for nn, ll in enumerate(self.lambdas):
            # print '\n------------------------------'
            pp = self.evalPencil(cvec) - ll * np.eye(MM)

            qq, rr, pp = la.qr(pp, pivoting=True)
            rmat = rr[:MM-1, :MM-1]
            rvec = rr[:MM-1, -1]
            rvar = rr[-1, -1]

            ptc = self.pencil[1:] - ll * np.eye(MM)[None,:,:]
            ptc = ptc[:,:,pp]
            tmat = np.zeros((KK-1, MM-1))
            tvec = np.zeros(KK-1)
            for kk, ppc in enumerate(ptc):
            # for kk in range(self.K-1):
                # ppc_ = self.pencil[kk+1] - ll * np.eye(MM)
                # ppc_ = ppc_[:,pp]

                # print 'ppc1 =', ppc
                # print 'ppc2 =', ppc_

                tt = np.dot(qq.T, ppc)
                tmat[kk] = tt[-1,:MM-1]
                # tmat[kk] = tt[:MM-1,-1]
                tvec[kk] = tt[-1, -1]

            self.nm.addCollection(rmat, rvec, rvar, tmat, tvec)

    def updateC(self, cvec, inplace=True, pprint=False):
        self._evalQRs(cvec)
        # Jacobian and Fvec
        fvec = self.nm.getFC()
        jfmat = self.nm.getJFC()

        print('!! Jf =')
        print(jfmat)

        jfaux = np.dot(jfmat.T, jfmat)
        # compute next C vec
        delta_c = la.solve(jfaux, -np.dot(jfmat.T, fvec))
        if pprint:
            print('>> norm(Delta C) =', la.norm(delta_c))
        if inplace:
            cvec += delta_c
            return cvec
        else:
            return cvec + delta_c

    def __call__(self, c0, eps=1e-5, max_iter=20, return_residual=True, pprint=False):
        rr = 10.*eps
        kk = 0
        cc = c0
        while kk < max_iter and rr > eps:
            self.updateC(cc, pprint=pprint)
            rr = self.nm.calcResidual()
            kk += 1
            if pprint:
                print('\n---- Iter no. %d ----'%kk)
                print('>> residual =', rr)
                print('>> c =', cc)

        if pprint:
            print('\n---- Final eigenvalues ----')
            eig, _ = la.eig(self.evalPencil(cc))
            print('>> eig =', eig)

        if return_residual:
            return cc, rr
        else:
            return cc

    def minimizeResiduals(self, c0, return_residual=True, pprint=True):

        def fun_aux(cc):
            self._evalQRs(cc)
            return self.nm.calcResidual()

        # res = so.minimize(fun_aux, c0, method='L-BFGS-B', options={'ftol': 1e-15, 'gtol': 1e-15})
        res = so.minimize(fun_aux, c0, method='TNC')
        c_sol = res['x']
        rr = res['fun']

        # print res

        if return_residual:
            return c_sol, rr
        else:
            return c_sol


class TestIEPSolver(object):
    def __init__(self, n_size=4, n_mat=6):
        self.n_size, self.n_mat = n_size, n_mat
        # define the matrix pensil
        pencil = [np.random.rand(n_size, n_size) for _ in range(n_mat)]
        pencil = np.array([p.T + p for p in pencil])
        self.ieps = IEPSolver(pencil)
        # compute eigenvalues
        self.c_orig = np.linspace(1., float(n_mat-1), n_mat-1)
        mat_orig = self.ieps.evalPencil(self.c_orig)
        lambdas_orig, _ = la.eig(mat_orig)
        self.lambdas_orig = lambdas_orig.real
        # set eigenvalues
        self.ieps.initLambdas(self.lambdas_orig[2:])

        # self.c_orig =np.array([5.,2.,7.,1.,3.])

    def testResiduals(self):
        # compute residuals and test that they are 0
        self.ieps._evalQRs(self.c_orig)
        assert np.allclose(self.ieps.nm.getFC(), np.zeros(self.ieps.N))

    def testJacobian(self, dc_over_c=0.01):
        ii = np.random.randint(self.n_mat-1)
        dc_arr = np.array([self.c_orig[ii] * dc_over_c if jj == ii else 0. for jj in range(self.n_mat-1)])
        c_1 = self.c_orig + dc_arr
        # evaluate residuals & jacobian
        self.ieps._evalQRs(self.c_orig)
        f0 = self.ieps.nm.getFC()
        jf0 = self.ieps.nm.getJFC()
        print('jf0 =\n', jf0)
        self.ieps._evalQRs(c_1)
        f1 = self.ieps.nm.getFC()
        print('\n!!!')
        print(ii)
        print(f0)
        print(f1)
        print('!!!\n')
        # evaluate approximate jacobian
        jf_col = (f1 - f0) / dc_arr[ii]
        # check if correct
        assert jf0.shape == (self.ieps.N, self.ieps.K-1)
        print(jf0[:,ii])
        print(jf_col)
        assert np.allclose(jf_col, jf0[:,ii])

    def plotContinuity(self, dc_range=np.linspace(-1.,1.,100)):
        ii = np.random.randint(self.n_mat-1)
        rvar_range = np.zeros((dc_range.shape[0], self.n_size))
        for jj, dc in enumerate(dc_range):
            dc_arr = np.array([self.c_orig[ii] * dc if kk == ii else 0. for kk in range(self.n_mat-1)])
            c_1 = self.c_orig + dc_arr

            self.ieps._evalQRs(c_1)
            rvar_range[jj] = self.ieps.nm.getFC()

        # for kk in range(self.n_size):
        #     pl.plot(dc_range, rvar_range[:,kk])
        # pl.show()

    def plotJacobian(self, dc_range=np.linspace(-1.,1.,100)):
        delta_c = dc_range[1] - dc_range[0]
        ii = np.random.randint(self.n_mat-1)
        rvar_range = np.zeros((dc_range.shape[0], self.n_size))
        drvar_range = np.zeros((dc_range.shape[0], self.n_size))
        for jj, dc in enumerate(dc_range):
            dc_arr = np.array([self.c_orig[ii] * dc if kk == ii else 0. for kk in range(self.n_mat-1)])
            c_1 = self.c_orig + dc_arr

            self.ieps._evalQRs(c_1)
            rvar_range[jj] = self.ieps.nm.getFC()
            drvar_range[jj] = self.ieps.nm.getJFC()[:,ii]

        dc_ = (dc_range[1:] + dc_range[:-1]) / 2.
        drvar_approx = (rvar_range[1:,:] - rvar_range[:-1,:]) / delta_c


        # # from datarep.matplotlibsettings import *
        # for kk in range(self.n_size):
        #     pl.plot(dc_, drvar_approx[:,kk], c=colours[kk%len(colours)])
        #     pl.plot(dc_range, drvar_range[:,kk], lw=2, ls='--', c=colours[kk%len(colours)])
        # pl.show()


    def testSolver(self, eps=1.):

        print('\nc_0 =', self.c_orig)
        p0 = self.ieps.evalPencil(self.c_orig)
        eig0, _ = la.eig(p0)
        print('eig_0 =', eig0.real)

        self.ieps._evalQRs(self.c_orig)
        r0 = self.ieps.nm.getFC()
        print('r_0 =', r0)


        c_perturb = self.c_orig + eps * np.random.randn(*self.c_orig.shape)
        print('\nc_1 =', c_perturb)

        p1 = self.ieps.evalPencil(c_perturb)
        eig1, _ = la.eig(p1)
        print('eig_1 =', eig1.real)

        self.ieps._evalQRs(c_perturb)
        r1 = self.ieps.nm.getFC()
        print('r_1 =', r1)

        # c_fit = ieps(c_perturb, pprint=True)
        c_fit, r_fit = self.ieps.minimizeResiduals(c_perturb, pprint=True)


        print('\nc_fit =', c_fit)
        p_fit = self.ieps.evalPencil(c_fit)
        eig2, _ = la.eig(p_fit)
        print('eig_fit =', eig2.real)

        self.ieps._evalQRs(c_fit)
        r2 = self.ieps.nm.getFC()
        print('r_fit =', r2)

if __name__ == '__main__':
    ts = TestIEPSolver()

    # ts.testResiduals()
    # ts.testJacobian()

    ts.testSolver()

    # ts.plotContinuity()
    # ts.plotJacobian()




"""
writes mod files for all channels defined in ``neat.channels.channelcollection``
"""
import os

from . import channelcollection

print('--> Writing mod files')
for name, channel_class in list(channelcollection.__dict__.items()):
    if isinstance(channel_class, type) and name != 'IonChannel' and name != '_func':
        chan = channel_class()
        chan.writeModFile(os.path.join(os.path.dirname(__file__),
                                       '../tools/simtools/neuron/mech'))
        del chan



"""
File contains:

    - `neat.SOVNode`
    - `neat.SomaSOVNode`
    - `neat.SOVTree`

Author: W. Wybo
"""

import numpy as np

import itertools
import copy

from . import morphtree
from .morphtree import MorphLoc
from .phystree import PhysNode, PhysTree
from .netree import NETNode, NET, Kernel

from ..tools.fittools import zerofinding as zf
from ..tools.fittools import histogramsegmentation as hs


def _consecutive(data, stepsize=1):
    return np.split(data, np.where(np.diff(data) != stepsize)[0]+1)


class SOVNode(PhysNode):
    """
    Node that defines functions and stores quantities to implement separation
    of variables calculation (Major, 1993)
    """
    def __init__(self, index, p3d=None):
        super().__init__(index, p3d)

    def _setSOV(self, channel_storage, tau_0=0.02):
        self.counter = 0
        # segment parameters
        self.g_m        = self.getGTot(channel_storage) # uS/cm^2
        # parameters for SOV approach
        self.R_sov      = self.R * 1e-4 # convert um to cm
        self.L_sov      = self.L * 1e-4 # convert um to cm
        self.tau_m      = self.c_m / self.g_m # s
        self.eps_m      = self.tau_m / tau_0
        self.lambda_m   = np.sqrt(self.R_sov / (2.*self.g_m*self.r_a)) # cm
        self.tau_0      = tau_0 # s
        self.z_a        = self.r_a / (np.pi * self.R_sov**2) # MOhm/cm
        self.g_inf_m    = 1. / (self.z_a * self.lambda_m) # uS
        # # segment amplitude information
        self.kappa_m    = np.NaN
        self.mu_vals_m  = np.NaN
        self.q_vals_m   = np.NaN

    def q_m(self, x):
        return np.sqrt(self.eps_m*x**2 - 1.)

    def dq_dp_m(self, x):
        return -self.tau_m / (2.*self.q_m(x))

    def mu_m(self, x):
        cns = self.child_nodes
        if len(cns) == 0:
            return self.z_a * self.lambda_m / self.q_m(x) * self.g_shunt
        else:
            x = zf._to_complex(x)
            mu_ds = [cn.mu_m(x) for cn in cns]
            q_ds  = [cn.q_m(x) for cn in cns]
            return ( self.g_shunt - np.sum([ \
                            cn.g_inf_m*q_ds[i] * \
                            ( 1.-mu_ds[i]/np.tan(q_ds[i]*cn.L_sov/cn.lambda_m) ) / ( 1./np.tan(q_ds[i]*cn.L_sov/cn.lambda_m)+mu_ds[i]) \
                            for i, cn in enumerate(cns)], 0) ) / (self.g_inf_m * self.q_m(x))

    def dmu_dp_m(self, x):
        cns = self.child_nodes
        if len(cns) == 0:
            return -self.dq_dp_m(x) * self.mu_m(x) / self.q_m(x)
        else:
            x = zf._to_complex(x)
            mu_ds = [cn.mu_m(x) for cn in cns]
            dmu_dp_ds = [cn.dmu_dp_m(x) for cn in cns]
            q_ds  = [cn.q_m(x) for cn in cns]
            dq_dp_ds = [cn.dq_dp_m(x) for cn in cns]
            return (-self.dq_dp_m(x) * self.mu_m(x) - np.sum([ cn.g_inf_m * ( \
                            dq_dp_ds[i] * \
                            ( 1.-mu_ds[i]/np.tan(q_ds[i]*cn.L_sov/cn.lambda_m) ) / ( 1./np.tan(q_ds[i]*cn.L_sov/cn.lambda_m)+mu_ds[i]) + \
                            q_ds[i] * \
                            ( (1.+mu_ds[i]**2) * dq_dp_ds[i] * cn.L_sov/cn.lambda_m - dmu_dp_ds[i] ) / \
                            ( np.cos(q_ds[i]*cn.L_sov/cn.lambda_m) + mu_ds[i]*np.sin(q_ds[i]*cn.L_sov/cn.lambda_m) )**2 \
                            ) for i, cn in enumerate(cns)], 0) / self.g_inf_m ) / self.q_m(x)

    def _setKappaFactors(self, xzeros):
        xzeros = zf._to_complex(xzeros)
        self.kappa_m = self.parent_node.kappa_m / \
                           (np.cos(self.q_m(xzeros)*self.L_sov/self.lambda_m) + \
                            self.mu_m(xzeros)*np.sin(self.q_m(xzeros)*self.L_sov/self.lambda_m))

    def _setMuVals(self, xzeros):
        xzeros = zf._to_complex(xzeros)
        self.mu_vals_m = self.mu_m(xzeros)

    def _setQVals(self, xzeros):
        xzeros = zf._to_complex(xzeros)
        self.q_vals_m = self.q_m(xzeros)

    def _findLocalPoles(self, maxspace_freq=500):
        poles = []
        pmultiplicities = []
        n = 0
        val = 0.
        while val < maxspace_freq:
            poles.append(np.sqrt((1.+val**2)/self.eps_m))
            if val == 0:
                pmultiplicities.append(.5)
            else:
                pmultiplicities.append(1.)
            n += 1
            val = n*np.pi * self.lambda_m / self.L_sov
        return poles, pmultiplicities

    def _setZerosPoles(self, maxspace_freq=500, pprint=False):
        cns = self.child_nodes
        # find the poles of (1 + mu*cot(qL/l))/(cot(qL/l) + mu)
        lpoles, lpmultiplicities = self._findLocalPoles(maxspace_freq)
        for cn in cns:
            lpoles.extend(cn.poles)
            lpmultiplicities.extend(cn.pmultiplicities)
        inds = np.argsort(lpoles)
        lpoles = np.array(lpoles)[inds]; lpmultiplicities = np.array(lpmultiplicities)[inds]
        # construct the function cot(qL/l) + mu
        f = lambda x: 1./np.tan(self.q_m(x)*self.L_sov/self.lambda_m) + self.mu_m(x)
        dfdx = lambda x: -2.*x * ( -(self.L_sov/self.lambda_m)/np.sin(self.q_m(x)*self.L_sov/self.lambda_m)**2 * self.dq_dp_m(x) +\
                             self.dmu_dp_m(x) ) / self.tau_0
        # find its zeros, this are the poles of the next level
        xval = 1.5/np.sqrt(self.eps_m)
        for cn in cns:
            c_eps_m = cn.eps_m
            xval_ = 1.5/np.sqrt(c_eps_m)
            if  xval_ > xval:
                xval = xval_
        if np.abs(f(xval)) < 1e-20:
            xval = (xval+lpoles[1])/2.
        if pprint:
            print('')
            print('xval: ', xval)
        # find zeros larger than xval
        if pprint: print('finding real poles')
        PF = zf.poleFinder(fun=f, dfun=dfdx, global_poles={'poles': lpoles, 'pmultiplicities': lpmultiplicities})
        poles, pmultiplicities = PF.find_real_zeros(vmin=xval)
        # find the first zero
        if pprint: print('finding first pole')
        p1 = []; pm1 = []
        zf.find_zeros_on_segment(p1, pm1, 0., xval, f, dfdx, lpoles, lpmultiplicities, pprint=pprint)
        self.poles = np.concatenate((p1, poles)).real; self.pmultiplicities = np.concatenate((pm1, pmultiplicities)).real


class SomaSOVNode(SOVNode):
    """
    Subclass of SOVNode to threat the special case of the soma

    The following member functions are not supposed to work properly,
    calling them may result in errors:
    `neat.SOVNode._setKappaFactors()`
    `neat.SOVNode._setMuVals()`
    `neat.SOVNode._setQVals()`
    `neat.SOVNode._findLocalPoles()`
    """
    def __init__(self, index, p3d=None):
        super().__init__(index, p3d)

    def _setSOV(self, channel_storage, tau_0=0.02):
        self.counter = 0
        # convert to cm
        self.R_sov      = self.R * 1e-4 # convert um to cm
        self.L_sov      = self.L * 1e-4 # convert um to cm
        # surface
        self.A = 4.0*np.pi*self.R_sov**2 # cm^2
        # total conductance
        self.g_m        = self.getGTot(channel_storage=channel_storage) # uS/cm^2
        # parameters for the SOV approach
        self.tau_m      = self.c_m / self.g_m # s
        self.eps_m      = self.tau_m / tau_0 # ns
        self.g_s        = self.g_m*self.A + self.g_shunt # uS
        self.c_s        = self.c_m*self.A # uF
        self.tau_0      = tau_0  # s
        # segment amplitude factors
        self.kappa_m = 1.

    def f_transc(self, x):
        cns = self.child_nodes
        x = zf._to_complex(x)
        mu_ds = [cn.mu_m(x) for cn in cns]
        q_ds  = [cn.q_m(x) for cn in cns]
        return self.g_s * (1.-self.eps_m*x**2) - np.sum([ \
                     cn.g_inf_m*q_ds[i] * \
                    ( 1.-mu_ds[i]/np.tan(q_ds[i]*cn.L_sov/cn.lambda_m) ) / ( 1./np.tan(q_ds[i]*cn.L_sov/cn.lambda_m)+mu_ds[i]) \
                    for i, cn in enumerate(cns)], 0)

    def dN_dp(self, x):
        cns = self.child_nodes
        x = zf._to_complex(x)
        mu_ds = [cn.mu_m(x) for cn in cns]
        dmu_dp_ds = [cn.dmu_dp_m(x) for cn in cns]
        q_ds  = [cn.q_m(x) for cn in cns]
        dq_dp_ds = [cn.dq_dp_m(x) for cn in cns]
        return self.c_s - np.sum([ cn.g_inf_m * ( \
                            dq_dp_ds[i] * \
                            ( 1.-mu_ds[i]/np.tan(q_ds[i]*cn.L_sov/cn.lambda_m) ) / ( 1./np.tan(q_ds[i]*cn.L_sov/cn.lambda_m)+mu_ds[i]) + \
                            q_ds[i] * \
                            ( (1.+mu_ds[i]**2) * dq_dp_ds[i] * cn.L_sov/cn.lambda_m - dmu_dp_ds[i] ) / \
                            ( np.cos(q_ds[i]*cn.L_sov/cn.lambda_m) + mu_ds[i]*np.sin(q_ds[i]*cn.L_sov/cn.lambda_m) )**2 \
                            ) for i, cn in enumerate(cns)], 0)

    def _setZerosPoles(self, maxspace_freq=500, pprint=False):
        # find the poles of cot(qL/l) + mu
        lpoles = []; lpmultiplicities = []
        for cn in self.child_nodes:
            lpoles.extend(cn.poles)
            lpmultiplicities.extend(cn.pmultiplicities)
        inds = np.argsort(lpoles)
        lpoles = np.array(lpoles)[inds]; lpmultiplicities = np.array(lpmultiplicities)[inds]
        # construct the function cot(qL/l) + mu
        f = lambda x: self.f_transc(x)
        dfdx = lambda x: -2.*x  * self.dN_dp(x) / self.tau_0
        # find its zeros, this are the inverse timescales of the model
        xval = 1.5/np.sqrt(self.eps_m)
        for cn in self.child_nodes:
            c_eps_m = cn.eps_m
            xval_ = 1.5/np.sqrt(c_eps_m)
            if  xval_ > xval:
                xval = xval_
        if np.abs(f(xval)) < 1e-20:
            xval = (xval+lpoles[1])/2.
        if pprint:
            print('xval: ', xval)
        # find zeros larger than xval
        PF = zf.poleFinder(fun=f, dfun=dfdx, global_poles={'poles': lpoles, 'pmultiplicities': lpmultiplicities})
        zeros, multiplicities = PF.find_real_zeros(vmin=xval)
        # find the first zero
        z1 = []; zm1 = []
        zf.find_zeros_on_segment(z1, zm1, 0., xval, f, dfdx, lpoles, lpmultiplicities, pprint=pprint)
        self.zeros = np.concatenate((z1, zeros)).real; self.zmultiplicities = np.concatenate((zm1, multiplicities)).real
        self.prefactors = self.dN_dp(self.zeros).real


class SOVTree(PhysTree):
    """
    Class that computes the separation of variables time scales and spatial
    mode functions for a given morphology and electrical parameter set. Employs
    the algorithm by (Major, 1994). This three defines a special
    `neat.SomaSOVNode` on as a derived class from `neat.SOVNode` as some
    functions required for SOV calculation are different and thus overwritten.

    The SOV calculation proceeds on the computational tree (see docstring of
    `neat.MorphNode`). Thus it makes no sense to look for sov quantities in the
    original tree.
    """
    def __init__(self, file_n=None, types=[1,3,4]):
        super().__init__(file_n=file_n, types=types)

    def _createCorrespondingNode(self, node_index, p3d=None):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        if node_index == 1:
            return SomaSOVNode(node_index, p3d=p3d)
        else:
            return SOVNode(node_index, p3d=p3d)

    @morphtree.computationalTreetypeDecorator
    def getSOVMatrices(self, locarg):
        """
        returns the alphas, the reciprocals of the mode time scales [1/ms]
        as well as the spatial functions evaluated at ``locs``

        Parameters
        ----------
            locarg: see :func:`neat.MorphTree._parseLocArg()`
                the locations at which to evaluate the SOV matrices

        Returns
        -------
            alphas: np.ndarray of complex (ndim = 1)
                the reciprocals of mode time-scales (kHz)
            gammas: np.ndarray of complex (ndim = 2)
                the spatial function associated with each mode, evaluated at
                each locations. Dimension 0 is number of modes and dimension 1
                number of locations
        """
        locs = self._parseLocArg(locarg)
        if len(self) > 1:
            # set up the matrices
            zeros      = self.root.zeros
            prefactors = self.root.prefactors
            alphas     = zeros**2 / (self.tau_0*1e3)
            gammas     = np.zeros((len(alphas), len(locs)), dtype=complex)
            # fill the matrix of prefactors
            for ii, loc in enumerate(locs):
                if loc['node'] == 1:
                    x = 0.
                    node = self.root.child_nodes[0]
                else:
                    x = loc['x']
                    node = self[loc['node']]
                # fill a column of the matrix, corresponding to current loc
                gammas[:, ii] = node.kappa_m * \
                   (np.cos(node.q_vals_m*(1.-x)*node.L_sov/node.lambda_m) + \
                   node.mu_vals_m * np.sin(node.q_vals_m*(1.-x)*node.L_sov/node.lambda_m)) / \
                   np.sqrt(prefactors*1e3)
        else:
            alphas = np.array([1e-3 / self.root.tau_m])
            gammas = np.array([[np.sqrt(alphas[0] / self.root.g_s)]])

        # return the matrices
        return alphas, gammas

    @morphtree.computationalTreetypeDecorator
    def calcSOVEquations(self, maxspace_freq=500., pprint=False):
        """
        Calculate the timescales and spatial functions of the separation of
        variables approach, using the algorithm by (Major, 1993).

        The (reciprocals) of the timescales (i.e. the roots of the transcendental
        equation) are stored in the somanode.
        The spatial factors are stored in each (computational) node.

        Parameters
        ----------
        maxspace_freq: float (default is 500)
            roughly corresponds to the maximal spatial frequency of the
            smallest time-scale mode
        """
        self.tau_0 = np.pi#1.
        for node in self: node._setSOV(self.channel_storage, tau_0=self.tau_0)
        if len(self) > 1:
            # start the recursion through the tree
            self._SOVFromLeaf(self.leafs[0], self.leafs[1:],
                                maxspace_freq=maxspace_freq, pprint=pprint)
            # zeros are now found, set the kappa factors
            zeros = self.root.zeros
            self._SOVFromRoot(self.root, zeros)
            # clean
            for node in self: node.counter = 0
        else:
            self[1]._setSOV(self.channel_storage, tau_0=self.tau_0)

    def _SOVFromLeaf(self, node, leafs, count=0,
                        maxspace_freq=500., pprint=False):
        if pprint:
            print('Forward sweep: ' + str(node))
        pnode = node.parent_node
        # log how many times recursion has passed at node
        if not self.isLeaf(node):
            node.counter += 1
        # if the number of childnodes of node is equal to the amount of times
        # the recursion has passed node, the mu functions can be set. Otherwise
        # we start a new recursion at another leaf.
        if node.counter == len(node.child_nodes):
            # node._setMuFunctions()
            node._setZerosPoles(maxspace_freq=maxspace_freq)
            if not self.isRoot(node):
                self._SOVFromLeaf(pnode, leafs, count=count+1,
                                maxspace_freq=maxspace_freq, pprint=pprint)
        elif len(leafs) > 0:
                self._SOVFromLeaf(leafs[0], leafs[1:], count=count+1,
                                maxspace_freq=maxspace_freq, pprint=pprint)

    def _SOVFromRoot(self, node, zeros):
        for cnode in node.child_nodes:
            cnode._setKappaFactors(zeros)
            cnode._setMuVals(zeros)
            cnode._setQVals(zeros)
            self._SOVFromRoot(cnode, zeros)

    def getModeImportance(self, locarg=None, sov_data=None,
                                importance_type='simple'):
        """
        Gives the overal importance of the SOV modes for a certain set of
        locations

        Parameters
        ----------
            locarg: None or list of locations
            sov_data: None or tuple of mode matrices
                One of the keyword arguments ``locarg`` or ``sov_data``
                must not be ``None``. If ``locarg`` is not ``None``, the importance
                is evaluated at these locations (see
                :func:`neat.MorphTree._parseLocArg`).
                If ``sov_data`` is not ``None``, it is a tuple of a vector of
                the reciprocals of the mode timescales and a matrix with the
                corresponding spatial mode functions.
            importance_type: string ('relative' or 'absolute')
                when 'absolute', returns an absolute measure of the importance,
                when 'relative', normalizes so that maximum importance is one.
                Defaults to 'relative'.

        Returns
        -------
            np.ndarray (ndim = 1)
                the importances associated with each mode for the provided set
                of locations
        """
        if locarg is not None:
            locs = self._parseLocArg(locarg)
            alphas, gammas = self.getSOVMatrices(locs)
        elif sov_data is not None:
            alphas = sov_data[0]
            gammas = sov_data[1]
        else:
            raise IOError('One of the kwargs `locarg` or `sov_data` must not be ``None``')

        if importance_type == 'simple':
            absolute_importance = np.sum(np.abs(gammas), 1) / np.abs(alphas)
        elif importance_type == 'full':
            absolute_importance = np.zeros(len(alphas))
            for kk, (alpha, phivec) in enumerate(zip(alphas, gammas)):
                absolute_importance[kk] = np.sqrt(np.sum(np.abs(np.dot(phivec[:,None], phivec[None,:]))) / np.abs(alpha))
        else:
            raise ValueError('`importance_type` argument can be \'simple\' or \
                              \'full\'')

        return absolute_importance / np.max(absolute_importance)

    def getImportantModes(self, locarg=None, sov_data=None,
                                eps=1e-4, sort_type='timescale',
                                return_importance=False):
        """
        Returns the most importand eigenmodes (those whose importance is above
        the threshold defined by `eps`)

        Parameters
        ----------
            locarg: None or list of locations
            sov_data: None or tuple of mode matrices
                One of the keyword arguments ``locarg`` or ``sov_data``
                must not be ``None``. If ``locarg`` is not ``None``, the importance
                is evaluated at these locations (see
                :func:`neat.MorphTree._parseLocArg`).
                If ``sov_data`` is not ``None``, it is a tuple of a vector of
                the reciprocals of the mode timescales and a matrix with the
                corresponding spatial mode functions.
            eps: float
                the cutoff threshold in relative importance below which modes
                are truncated
            sort_type: string ('timescale' or 'importance')
                specifies in which order the modes are returned. If 'timescale',
                modes are sorted in order of decreasing time-scale, if
                'importance', modes are sorted in order of decreasing importance.
            return_importance: bool
                if ``True``, returns the importance metric associated with each
                mode

        Returns
        -------
            alphas: np.ndarray of complex (ndim = 1)
                the reciprocals of mode time-scales ``[kHz]``
            gammas: np.ndarray of complex (ndim = 2)
                the spatial function associated with each mode, evaluated at
                each locations. Dimension 0 is number of modes and dimension 1
                number of locations
            importance: np.ndarray (`shape` matches `alphas`, only if `return_importance` is ``True``)
                value of importance metric for each mode
        """
        if locarg is not None:
            locs = self._parseLocArg(locarg)
            alphas, gammas = self.getSOVMatrices(locs)
        elif sov_data is not None:
            alphas = sov_data[0]
            gammas = sov_data[1]
        else:
            raise IOError('One of the kwargs `locarg` or `sov_data` must not be ``None``')
        importance = self.getModeImportance(sov_data=(alphas, gammas), importance_type='simple')
        inds = np.where(importance > eps)[0]
        # only modes above importance cutoff
        alphas, gammas, importance = alphas[inds], gammas[inds,:], importance[inds]
        if sort_type == 'timescale':
            inds_sort = np.argsort(np.abs(alphas))
        elif sort_type == 'importance':
            inds_sort = np.argsort(importance)[::-1]
        else:
            raise ValueError('`sort_type` argument can be \'timescale\' or \
                              \'importance\'')
        if return_importance:
            return alphas[inds_sort], gammas[inds_sort,:], importance[inds_sort]
        else:
            return alphas[inds_sort], gammas[inds_sort,:]

    def calcImpedanceMatrix(self, locarg=None, sov_data=None, name=None,
                                  eps=1e-4, mem_limit=500, freqs=None):
        """
        Compute the impedance matrix for a set of locations

        Parameters
        ----------
            locarg: None or list of locations
            sov_data: None or tuple of mode matrices
                One of the keyword arguments ``locarg`` or ``sov_data``
                must not be ``None``. If ``locarg`` is not ``None``, the importance
                is evaluated at these locations (see
                :func:`neat.MorphTree._parseLocArg`).
                If ``sov_data`` is not ``None``, it is a tuple of a vector of
                the reciprocals of the mode timescales and a matrix with the
                corresponding spatial mode functions.
            eps: float
                the cutoff threshold in relative importance below which modes
                are truncated
            mem_limit: int
                parameter governs whether the fast (but memory intense) method
                or the slow method is used
            freqs: np.ndarray of complex or None (default)
                if ``None``, returns the steady state impedance matrix, if
                a array of complex numbers, returns the impedance matrix for
                each Fourrier frequency in the array

        Returns
        -------
            np.ndarray of floats (ndim = 2 or 3)
                the impedance matrix, steady state if `freqs` is ``None``, the
                frequency dependent impedance matrix if `freqs` is given, with
                the frequency dependence at the first dimension ``[MOhm ]``
        """
        if locarg is not None:
            locs = self._parseLocArg(locarg)
            alphas, gammas = self.getSOVMatrices(locs)
        elif sov_data is not None:
            alphas = sov_data[0]
            gammas = sov_data[1]
        else:
            raise IOError('One of the kwargs `locarg` or `sov_data` must not be ``None``')
        n_loc = gammas.shape[1]
        if freqs is None:
            # construct the 2d steady state matrix
            y_activation = 1. / alphas
            # compute the matrix, methods depends on memory limit
            if gammas.shape[1] < mem_limit and gammas.shape[0] < int(mem_limit/2.):
                z_mat = np.sum(gammas[:,:,np.newaxis] * \
                               gammas[:,np.newaxis,:] * \
                               y_activation[:,np.newaxis,np.newaxis], 0).real
            else:
                z_mat = np.zeros((n_loc, n_loc))
                for ii, jj in itertools.product(range(n_loc), range(n_loc)):
                    z_mat[ii,jj] = np.sum(gammas[:,ii] * \
                                          gammas[:,jj] * \
                                          y_activation).real
        else:
            # construct the 3d fourrier matrix
            y_activation = 1e3 / (alphas[np.newaxis,:]*1e3 + freqs[:,np.newaxis])
            z_mat = np.zeros((len(freqs), n_loc, n_loc), dtype=complex)
            for ii, jj in itertools.product(range(n_loc), range(n_loc)):
                z_mat[:,ii,jj] = np.sum(gammas[np.newaxis,:,ii] * \
                                        gammas[np.newaxis,:,jj] * \
                                        y_activation, 1)
        return z_mat

    def constructNET(self, dz=50., dx=10., eps=1e-4,
                        use_hist=False, add_lin_terms=True,
                        improve_input_impedance=False,
                        pprint=False):
        """
        Construct a Neural Evaluation Tree (NET) for this cell

        Parameters
        ----------
        dz: float
            the impedance step for the NET model derivation
        dx: float
            the distance step to evaluate the impedance matrix
        eps: float
            the cutoff threshold in relative importance below which modes
            are truncated
        use_hist: bool
            whether or not to use histogram segmentations to find well
            separated parts of the dendritic tree (such ass apical tree)
        add_lin_terms:
            take into account that the optained NET will be used in conjunction
            with linear terms

        Returns
        -------
        `neat.NETree`
            The neural evaluation tree (Wybo et al., 2019) associated with the
            morphology.
        """
        # create a set of location at which to evaluate the impedance matrix
        self.distributeLocsUniform(dx=dx, name='net eval')
        # compute the z_mat matrix
        alphas, gammas = self.getImportantModes(locarg='net eval', eps=eps)
        z_mat = self.calcImpedanceMatrix(sov_data=(alphas, gammas))
        # derive the NET
        net = NET()
        self._addLayerA(net, None,
                        z_mat, alphas, gammas,
                        0., 0, np.arange(len(self.getLocs('net eval'))),
                        dz=dz,
                        use_hist=use_hist, add_lin_terms=add_lin_terms,
                        pprint=pprint)
        net.setNewLocInds()
        if improve_input_impedance:
            self._improveInputImpedance(net, alphas, gammas)
        if add_lin_terms:
            lin_terms = self.computeLinTerms(net, sov_data=(alphas, gammas))
            return net, lin_terms
        else:
            return net

    def _addLayerA(self, net, pnode,
                        z_mat, alphas, gammas,
                        z_max_prev, z_ind_0, true_loc_inds,
                        dz=100.,
                        use_hist=True, add_lin_terms=False,
                        pprint=False):
        # create a histogram
        n_bin = 15
        z_hist = np.histogram(z_mat[0,:], n_bin, density=False)
        # find the histogram partition
        h_ftc = hs.histogramSegmentator(z_hist)
        s_inds, p_inds = h_ftc.partition_fine_to_coarse(eps=1.4)
        while len(s_inds) > 3:
            s_inds = np.delete(s_inds, 1)

        # identify the necessary node indices and kernel computation indices
        node_inds  = []
        kernel_inds = []
        min_inds = []
        for ii, si in enumerate(s_inds[:-1]):
            if si > 0:
                n_inds = np.where(z_mat[0,:] > z_hist[1][si+1])[0]
                k_inds = np.where(np.logical_and(
                                    z_mat[0,:] > z_hist[1][si+1],
                                    z_mat[0,:] <= z_hist[1][s_inds[ii+1]+1]))[0]
                min_ind = np.argmin(z_mat[0,k_inds])
                min_inds.append(min_ind)
            else:
                n_inds = np.where(z_mat[0,:] >= z_hist[1][0])[0]
                k_inds = np.where(np.logical_and(
                                    z_mat[0,:] >= z_hist[1][0],
                                    z_mat[0,:] <= z_hist[1][s_inds[ii+1]+1]))[0]
                min_ind = np.argmin(z_mat[0,k_inds])
                min_inds.append(min_ind)
            node_inds.append(n_inds)
            kernel_inds.append(k_inds)

        # add NET nodes to the NET tree
        for ii, n_inds in enumerate(node_inds):
            k_inds = kernel_inds[ii]
            if len(k_inds) != 0:
                if add_lin_terms:
                    # get the minimal kernel
                    gammas_avg = gammas[:,0] * \
                                 gammas[:,k_inds[min_inds[ii]]]
                else:
                    # get the average kernel
                    if len(k_inds) < 100000:
                        gammas_avg = np.mean(gammas[:,0:1] * \
                                             gammas[:,k_inds], 1)
                    else:
                        inds_ = np.random.choice(k_inds, size=100000)
                        gammas_avg = np.mean(gammas[:,0:1] * \
                                             gammas[:,inds_], 1)
                z_avg_approx = np.sum(gammas_avg / alphas).real
                self._subtractParentKernels(gammas_avg, pnode)
                # add a node to the tree
                node = NETNode(len(net), true_loc_inds[n_inds],
                                z_kernel=(alphas, gammas_avg))
                if pnode != None:
                    net.addNodeWithParent(node, pnode)
                else:
                    net.root = node
                # set new pnode
                pnode = node

                # print stuff
                if pprint:
                    print(node)
                    print('n_loc =', len(node.loc_inds))
                    print('(locind0, size) = ', (k_inds[0], z_mat.shape[0]))
                    print('')

                if k_inds[0] == 0:
                    # start new branches, split where they originate from soma by
                    # checking where input impedance is close to somatic transfer
                    # impedance
                    z_max = z_hist[1][s_inds[ii+1]]
                    # check where new dendritic branches start
                    z_diag = z_mat[k_inds, k_inds]
                    z_x0   = z_mat[k_inds, 0]
                    b_inds = np.where(np.abs(z_diag - z_x0) < dz / 2.)[0][1:].tolist()
                    if len(b_inds) > 0:
                        if b_inds[0] != 1:
                            b_inds = [1] + b_inds
                        kk = len(b_inds)-1
                        while kk > 0:
                            if b_inds[kk]-1 == b_inds[kk-1]:
                                del b_inds[kk]
                            kk -= 1
                    else:
                        b_inds = [1]
                    for jj, i0 in enumerate(b_inds):
                        # make new z_mat matrix
                        i1 = len(k_inds) if i0 == b_inds[-1] else b_inds[jj+1]
                        inds = np.meshgrid(k_inds[i0:i1], k_inds[i0:i1],
                                           indexing='ij')
                        z_mat_new = copy.deepcopy(z_mat[inds[0], inds[1]])
                        # move further in the tree
                        self._addLayerB(net, node,
                                    z_mat_new, alphas, gammas,
                                    z_max, k_inds[i0:i1], dz=dz,
                                    use_hist=use_hist, add_lin_terms=add_lin_terms)
                else:
                    # make new z_mat matrix
                    k_seqs = _consecutive(k_inds)
                    if pprint:
                        print('\n>>> consecutive')
                        print('nseq:', len(k_seqs))
                        for k_seq in k_seqs: print('sequence:', k_seq)
                    for k_seq in k_seqs:
                        inds = np.meshgrid(k_seq, k_seq, indexing='ij')
                        z_mat_new = copy.deepcopy(z_mat[inds[0], inds[1]])
                        z_max = z_mat[0,0]+1
                        # move further in the tree
                        self._addLayerB(net, node,
                                z_mat_new, alphas, gammas,
                                z_max, k_seq, dz=dz, pprint=pprint,
                                use_hist=use_hist, add_lin_terms=add_lin_terms)

    def _addLayerB(self, net, pnode,
                z_mat, alphas, gammas,
                z_max_prev, true_loc_inds, dz=100.,
                use_hist=True, pprint=False, add_lin_terms=False):
        # print stuff
        if pprint:
            print('>>> node index = ', node._index)
            if pnode != None:
                print('parent index = ', pnode._index)
            else:
                print('start')
        # get the diagonal
        z_diag = np.diag(z_mat)

        if true_loc_inds[0] == 0 and z_mat[0,0] > z_max_prev:
            n_bins = 'soma'
            z_max = z_mat[0,0] + 1.
            z_min = z_max_prev
        else:
            # histogram GF
            n_bins = max(int(z_mat.size/50.),
                         int((np.max(z_mat) - np.min(z_mat))/dz))
            if n_bins > 1:
                if np.all(np.diff(z_diag) > 0):
                    z_min = z_max_prev
                    z_max = z_min + dz
                    if pprint: print('--> +', dz)
                elif use_hist:
                    z_hist = np.histogram(z_mat.flatten(), n_bins, density=False)
                    # find the histogram partition
                    h_ftc = hs.histogramSegmentator(z_hist)
                    s_ind, p_ind = h_ftc.partition_fine_to_coarse()

                    # get the new min max values
                    z_histx = z_hist[1]
                    z_min = z_max_prev
                    z_max = z_histx[s_ind[1]]
                    ii = 1
                    while np.min(z_diag) > z_histx[s_ind[ii]]:
                        ii += 1
                        z_max = z_histx[s_ind[ii]]
                    ii = np.argmax(z_hist[0][s_ind[0]:s_ind[ii]])
                    z_avg = z_hist[0][ii]
                    if z_max - z_min > dz:
                        z_max = z_min + dz
                    if pprint: print('--> hist: +', str(z_max - z_min))
                else:
                    z_min = z_max_prev
                    z_max = z_min + dz
                    if pprint: print('--> +', dz)
            else:
                z_min = z_max_prev
                z_max = np.max(z_mat)
                if pprint: print('--> all: +', str(z_max - z_min))
        d_inds = np.where(z_diag <= z_max+1e-15)[0]
        # make sure that there is at least one element in the layer
        while len(d_inds) == 0:
            z_max += dz
            d_inds = np.where(z_diag <= z_max+1e-15)[0]

        # identify different domains
        if add_lin_terms and true_loc_inds[0] == 0:
            t0 = np.array([1]); t1 = np.array([len(z_diag)])
        else:
            t0 = np.where(np.logical_and(z_diag[:-1] < z_max+1e-15,
                                         z_diag[1:] >= z_max+1e-15))[0]
            if len(t0) > 0: t0 += 1
            if z_diag[0] >= z_max+1e-15:
                t0 = np.concatenate(([0], t0))
            t1 = np.where(np.logical_and(z_diag[:-1] >= z_max+1e-15,
                                         z_diag[1:] < z_max+1e-15))[0]
            if len(t1) > 0: t1 += 1
            if z_diag[-1] >= z_max+1e-15:
                t1 = np.concatenate((t1, [len(z_diag)]))

        # identify where the kernels are within the interval
        l_inds = np.where(z_mat <= z_max+1e-15)

        # get the average kernel
        if l_inds[0].size < 100000:
            gammas_avg = np.mean(gammas[:,true_loc_inds[l_inds[0]]] * \
                                 gammas[:,true_loc_inds[l_inds[1]]], 1)
        else:
            inds_ = np.random.randint(l_inds[0].size, size=100000)
            gammas_avg = np.mean(gammas[:,true_loc_inds[l_inds[0]][inds_]] * \
                                 gammas[:,true_loc_inds[l_inds[1]][inds_]], 1)
        self._subtractParentKernels(gammas_avg, pnode)

        # add a node to the tree
        node = NETNode(len(net), true_loc_inds, z_kernel=(alphas, gammas_avg))
        if pnode != None:
            net.addNodeWithParent(node, pnode)
        else:
            net.root = node

        if pprint:
            print('(locind0, size) = ', (true_loc_inds[0], z_mat.shape[0]))
            print('(zmin, zmax, n_bins) = ', (z_min, z_max, n_bins))
            print('')

        # move on to the next layers
        if len(d_inds) < len(z_diag):
            for jj, ind0 in enumerate(t0):
                ind1 = t1[jj]
                z_mat_new = copy.deepcopy(z_mat[ind0:ind1,ind0:ind1])
                true_loc_inds_new = true_loc_inds[ind0:ind1]
                self._addLayerB(net, node,
                            z_mat_new, alphas, gammas,
                            z_max, true_loc_inds_new, dz=dz,
                            use_hist=use_hist, pprint=pprint)

    def _subtractParentKernels(self, gammas, pnode):
        if pnode != None:
            gammas -= pnode.z_kernel['c']
            self._subtractParentKernels(gammas, pnode.parent_node)

    def _improveInputImpedance(self, net, alphas, gammas):
        nmaxind = np.max([n.index for n in net])
        for node in net:
            if len(node.loc_inds) == 1:
                # recompute the kernel of this single loc layer
                if node.parent_node is not None:
                    p_kernel = net.calcTotalKernel(node.parent_node)
                    p_k_c = p_kernel.c
                else:
                    p_k_c = np.zeros_like(gammas)
                gammas_real = gammas[:,node.loc_inds[0]]**2
                node.z_kernel.c = gammas_real - p_k_c
            elif len(node.newloc_inds) > 0:
                z_k_approx = net.calcTotalKernel(node)
                # add new input nodes for the nodes that don't have one
                for ind in node.newloc_inds:
                    nmaxind += 1
                    gammas_real = gammas[:,ind]**2
                    z_k_real = Kernel(dict(a=alphas, c=gammas_real))
                    # add node
                    newnode = NETNode(nmaxind, [ind], z_kernel=z_k_real-z_k_approx)
                    newnode.newloc_inds = [ind]
                    net.addNodeWithParent(newnode, node)
                # empty the new indices
                node.newloc_inds = []
        net.setNewLocInds()

    def computeLinTerms(self, net, sov_data=None, eps=1e-4):
        """
        Construct linear terms for `net` so that transfer impedance to soma is
        exactly matched

        Parameters
        ----------
        net: `neat.NETree`
            the neural evaluation tree (NET)
        sov_data: None or tuple of mode matrices
            If ``sov_data`` is not ``None``, it is a tuple of a vector of
            the reciprocals of the mode timescales and a matrix with the
            corresponding spatial mode functions.
        eps: float
            the cutoff threshold in relative importance below which modes
            are truncated

        Returns
        -------
        lin_terms: dict of {int: `neat.Kernel`}
            the kernels associated with linear terms of the NET, keys are
            indices of their corresponding location stored inder 'net eval'
        """
        if sov_data != None:
            alphas = sov_data[0]
            gammas = sov_data[1]
        else:
            alphas, gammas = self.getImportantModes(locarg='net eval', eps=eps)
        lin_terms = {}
        for ii, loc in enumerate(self.getLocs('net eval')):
            if not self.isRoot(self[loc['node']]):
                # create the true kernel
                z_k_true = Kernel((alphas, gammas[:,ii] * gammas[:,0]))
                # compute the NET approximation kernel
                z_k_net = net.getReducedTree([0, ii]).getRoot().z_kernel
                # compute the lin term
                lin_terms[ii] = z_k_true - z_k_net

        return lin_terms




"""
File contains:

    - :class:`Kernel`
    - :class:`NETNode`
    - :class:`NET`

Author: W. Wybo
"""


import numpy as np
import matplotlib.pyplot as pl

from .stree import STree, SNode

import copy


class Kernel(object):
    """
    Implements a kernel as a superposition of exponentials:

    .. math:: k(t) = \sum_n c_n e^{ - a_n t}

    Kernels can be added and subtracted, as this class overloads the __add__
    and __subtract__ functions.

    They can be evaluated as a function of time by calling the object with a
    time array.

    They can be evaluated in the Fourrier domain with `Kernel.ft`

    Parameters
    ----------
    kernel: dict, float, neat.Kernel, tuple or list
        If dict, has the form {'a': `np.array`, 'c': `np.array`}.
        If float, sets `c` single exponential prefactor and assumes `a` is 1 kHz.
        If `neat.Kernel`, copies the object.
        If tuple or list, sets 'a' as first element and 'c' as last element.

    Attributes
    ----------
    a: np.array of float or complex
        The exponential coefficients (kHz)
    c: np.array of float or complex
        The exponential prefactors
    """
    def __init__(self, kernel):
        # set kernel time scales and exponential prefactors
        if isinstance(kernel, dict):
            self.a = copy.deepcopy(kernel['a'])
            self.c = copy.deepcopy(kernel['c'])
        elif isinstance(kernel, float) or isinstance(kernel, int):
            self.a = np.array([1.])
            self.c = np.array([kernel]).astype(float)
        elif isinstance(kernel, Kernel):
            self.a = copy.deepcopy(kernel.a)
            self.c = copy.deepcopy(kernel.c)
        else:
            self.a = copy.deepcopy(kernel[0])
            self.c = copy.deepcopy(kernel[1])
        if isinstance(self.a, float):
            self.a = np.array([self.a])
        elif not isinstance(self.a, np.ndarray):
            self.a = np.array(self.a)
        if isinstance(self.c, float):
            self.c = np.array([self.c])
        elif not isinstance(self.c, np.ndarray):
            self.c = np.array(self.c)

    def __getitem__(self, ind):
        if ind == 0: return self.a
        elif ind == 1: return self.c
        elif ind == 'a': return self.a
        elif ind == 'c': return self.c
        elif ind == 'alphas': return self.a
        elif ind == 'gammas': return self.c
        else: raise IndexError('Index should be \'0\' or \'1\'')

    def __call__(self, t_arr):
        return np.dot(np.exp(-t_arr[:,np.newaxis] * self.a[np.newaxis,:]), \
                      self.c[:,np.newaxis]).flatten().real

    def __add__(self, kernel):
        if kernel.a.shape[0] == self.a.shape[0] and \
           np.allclose(kernel.a, self.a):
            a = copy.copy(self.a)
            c = kernel.c + self.c
        else:
            a = np.concatenate((self.a, kernel.a))
            c = np.concatenate((self.c, kernel.c))
        return Kernel((a, c))

    def __sub__(self, kernel):
        if kernel.a.shape[0] == self.a.shape[0] and \
           np.allclose(kernel.a, self.a):
            a = copy.copy(self.a)
            c = self.c - kernel.c
        else:
            a = np.concatenate((self.a, kernel.a))
            c = np.concatenate((self.c, -kernel.c))
        return Kernel((a, c))

    def getKBar(self):
        """
        The total surface under the kernel
        """
        return np.sum(self.c / self.a).real

    def setKBar(self, kk):
        raise AttributeError('`k_bar` is a read-only attribute, adjust attribute `c` ' + \
                             'by multiplying with a factor to change `k_bar`')
    k_bar = property(getKBar, setKBar)

    def __str__(self, as_timescale=False):
        if as_timescale:
            return 't = ' + np.array2string(1./self.a, precision=4, max_line_width=1000) + '\n' + \
                   'c = ' + np.array2string(self.c, precision=4, max_line_width=1000)
        else:
            return 'a = ' + np.array2string(self.a, precision=4, max_line_width=1000) + '\n' + \
                   'c = ' + np.array2string(self.c, precision=4, max_line_width=1000)

    def t(self, t_arr):
        """
        Evaluates the kernel in the time domain

        Parameters
        ----------
        t_arr: np.array of float
            the time array at which the kernel is evaluated

        Returns
        -------
        np.array of float
            the temporal kernel
        """
        return self(t_arr)

    def ft(self, s_arr):
        """
        Evaluates the kernel in the Fourrier domain

        Parameters
        ----------
        s_arr: np.array of complex
            The frequencies (Hz) at which the kernel is to be evaluated

        Returns
        -------
        np.array of complex
            The Fourrier transform of the kernel

        """
        return np.sum(self.c[:,None]*1e3 / (self.a[:,None]*1e3 + s_arr[None,:]), 0)


class NETNode(SNode):
    """
    Node associated with `neat.NET`.

    Attributes
    ----------
    loc_inds: list of int
        The inidices of locations which the node integrates
    newloc_inds: list of int
        The locations for which the node is the most local component to integrate
        them
    z_kernel: `neat.Kernel`
        The impedance kernel with which the node integrates inputs
    z_bar: float
        The steady state impedance associated with the impedance kernel
    """
    def __init__(self, index, loc_inds, newloc_inds=[], z_kernel=None):
        super().__init__(index)
        # location indices that node integrates
        self.loc_inds = loc_inds
        self.newloc_inds = newloc_inds
        # kernel associated with node
        self.z_kernel = z_kernel

    def __str__(self):
        if self.parent_node is not None:
            return 'NETNode ' + str(self.index) + \
                    ', loc inds: ' + str(self.loc_inds) + \
                    ', newloc inds: ' + str(self.newloc_inds) + \
                    ', parent: ' + str(self.parent_node.index) + \
                    ', z_bar (MOhm) = ' + str(self.z_bar)
        else:
            return 'NETNode ' + str(self.index) + \
                    ', loc inds: ' + str(self.loc_inds) + \
                    ', newloc inds: ' + str(self.newloc_inds) + \
                    ', parent: None' \
                    ', z_bar (MOhm) = ' + str(self.z_bar)

    def setZKernel(self, z_kernel):
        self._z_kernel = Kernel(z_kernel)

    def getZKernel(self):
        return self._z_kernel

    def getZ(self):
        return self._z_kernel.k_bar

    z_kernel = property(getZKernel, setZKernel)
    z_bar = property(getZ, setZKernel)

    def __contains__(self, loc_ind):
        return loc_ind in self.loc_inds

    def _setCompartmentData(self, node_list, z_root_list, z_comp_list, Iz=5.):
        node_inds  = [node.index for node in node_list if node != None]
        z_root = np.array(z_root_list)
        z_comp = np.array(z_comp_list)
        comp_inds = np.where(z_comp / z_root > Iz)[0]
        # store the relevant quantities
        self._z_root = z_root[comp_inds]
        self._z_comp = z_comp[comp_inds]
        self._node_inds = [node_inds[ind] for ind in comp_inds]

    def _setTentativeCompartments(self, comps):
        self._comps = comps

    def _setSharedRootInd(self, ind):
        self._root_ind = self._node_inds.index(ind)


class NET(STree):
    """
    Abstract tree class that implements the Neural Evaluation Tree
    (Wybo et al., 2019), representing the spatial voltage as a number of voltage
    components present at different spatial scales.
    """
    def __init__(self, root=None):
        super().__init__(root)

    def __str__(self):
        string = 'NET\n'
        for node in self:
            string += '  > ' + str(node) + '\n'
        return string

    def _createCorrespondingNode(self, node_index):
        """
        Creates a node with the given index corresponding to the tree class.

        Parameters
        ----------
            node_index: int
                index of the new node
        """
        return NETNode(node_index, [])

    def getLocInds(self, sroot=None):
        """
        Get the indices of the locations a subtree integrates

        Parameters
        ----------
        sroot: `neat.NETNode`, int or None
            Root of the subtree, or index of the root. If ``None``, subtree is
            the whole tree.

        Returns
        -------
        loc_inds: indices of locations
        """
        if isinstance(sroot, int):
            sroot = self[sroot]
        elif sroot is None:
            sroot = self.root
        return sroot.loc_inds

    def getLeafLocNode(self, loc_ind):
        """
        Get the node for which ``loc_ind`` is a new location

        Parameters
        ----------
        loc_ind: int
            index of the location

        Returns
        -------
        :obj:`NETNode`
        """
        for node in self:
            if loc_ind in node.newloc_inds:
                return node

    def setNewLocInds(self):
        """
        Set the new location indices in a tree
        """
        for node in self:
            cloc_inds = set()
            for cnode in node.child_nodes:
                cloc_inds = cloc_inds.union(set(cnode.loc_inds))
            node.newloc_inds = list(set(node.loc_inds) - cloc_inds)

    def getReducedTree(self, loc_inds, indexing='NET eval'):
        """
        Construct a reduced tree where only the locations index by ``loc_inds''
        are retained

        Parameters
        ----------
        loc_inds : iterable of ints
            the indices of the locations that are to be retained
        indexing : 'NET eval' or 'locs'
            if 'NET eval', indexing of ``NETNode.loc_inds`` will be taken to be the
            indices of locations for which the full NET is evaluated. Otherwise
            will be indices of the input ``loc_inds``
        """
        loc_inds_newtree = list({loc_ind for loc_ind in loc_inds \
                                         if loc_ind in self.root})
        if loc_inds_newtree:
            new_root = NETNode(0, loc_inds_newtree,
                                z_kernel=self.root.z_kernel)
            new_tree = NET(new_root)
            for cnode in self.root.child_nodes:
                if cnode is not None:
                    self._constructReducedTree(cnode, loc_inds_newtree,
                                                new_root, new_tree)
            new_tree.setNewLocInds()
            if indexing == 'NET eval':
                return new_tree
            else:
                for node in new_tree:
                    # node.loc_inds = [np.where(loc_inds == ind)[0][0] for ind in node.loc_inds]
                    # node.loc_inds = sum([np.where(loc_inds == ind)[0].tolist() for ind in set(node.loc_inds)], [])
                    node.loc_inds = sum([np.where(loc_inds == ind)[0].tolist() for ind in node.loc_inds], [])
                new_tree.setNewLocInds()
                return new_tree
        else:
            return None

    def _constructReducedTree(self, node, loc_inds, node_newtree, new_tree):
        loc_inds_subtree = list({loc_ind for loc_ind in loc_inds \
                                         if loc_ind in node})
        if len(loc_inds_subtree) > 0:
            if loc_inds_subtree == loc_inds:
                node_newtree.z_kernel += node.z_kernel
            else:
                newnode_newtree = NETNode(len(new_tree), loc_inds_subtree,
                                            z_kernel=node.z_kernel)
                new_tree.addNodeWithParent(newnode_newtree, node_newtree)
                node_newtree = newnode_newtree
            for cnode in node.child_nodes:
                if cnode is not None:
                    self._constructReducedTree(cnode, loc_inds_subtree,
                                                node_newtree, new_tree)

    # def matchInputImpedance(self, z_input):
    #     assert imp_mat.shape[0] == imp_mat.shape[1]
    #     assert imp_mat.shape[0] == len(self.root.loc_inds)
    #     for node in self:
    #         if self.isLeaf(node):
    #             if len(node.loc_inds) == 1:
    #                 p_imp = self.calcTotalImpedance(node.parent_node)
    #                 node.z_kernel.c *= (z_input[node.locs_inds[0]] - p_imp) / node.z_kernel.k_bar
    #             else:
    #                 for loc_ind in node.loc_inds:
    #                     new_node = NETNode(len(tree), [loc_ind])
    #                     self.addNodeWithParent

    def calcTotalImpedance(self, node):
        """
        Compute the total impedance associated with a node. I.e. the sum of all
        impedances on the path from node to root

        Parameters
        ----------
        node : :class:`SNode`

        Returns
        -------
        float
            total impedance
        """
        return np.sum([node_.z_bar for node_ in self.pathToRoot(node)])

    def calcTotalKernel(self, node):
        """
        Compute the total impedance kernel associated with a node. I.e. the sum
        of all impedance kernels on the path from node to root

        Parameters
        ----------
        node : :class:`SNode`

        Returns
        -------
        :class:`Kernel`
        """
        z_k = copy.deepcopy(node.z_kernel)
        if node.parent_node is not None:
            for pn in self.pathToRoot(node.parent_node):
                z_k += pn.z_kernel
        return z_k

    def calcIZ(self, loc_inds):
        """
        compute I_Z between any pair of locations in ``loc_inds``

        Parameters
        ----------
        loc_inds : iterable of ints
            the indices of locations between which I_Z has to be evaluated

        Returns
        -------
        float or dict of tuple : float
            Returns a float if the number of location indices is two, otherwise
            a dictionary with location pairs (smallest is listed first) as keys
            and I_Z values as values
        """
        Iz_dict = {}
        for ii, loc_ind0 in enumerate(loc_inds):
            for jj, loc_ind1 in enumerate(loc_inds):
                if jj < ii:
                    net_red = self.getReducedTree([loc_ind0, loc_ind1])
                    key = (loc_ind0, loc_ind1) if loc_ind0 < loc_ind1 \
                                               else (loc_ind1, loc_ind0)
                    n0 = net_red.getLeafLocNode(loc_ind0)
                    z0 = n0.z_bar if n0 != net_red.root else 0.
                    n1 = net_red.getLeafLocNode(loc_ind1)
                    z1 = n1.z_bar if n1 != net_red.root else 0.
                    Iz_dict[key] = (z0 + z1) / (2. * net_red.root.z_bar)
                else:
                    break
        if len(loc_inds) == 2:
            return list(Iz_dict.values())[0]
        else:
            return Iz_dict

    def calcIZMatrix(self):
        """
        compute the Iz matrix for all locations present in the tree

        Returns
        -------
        np.ndarray of float
            The Iz matrix
        """
        z_mat = self.calcImpedanceMatrix()
        z_in = np.diag(z_mat)
        return (z_in[:,np.newaxis] + z_in[np.newaxis,:]) / (2. * z_mat) - 1.

    def calcImpedanceMatrix(self):
        """
        Compute the impedance matrix approximation associated with the NET

        Returns
        -------
        np.ndarray (ndim = 2)
            the impedance matrix approximation
        """
        return self.calcImpMat()

    def calcImpMat(self):
        """
        Compute the impedance matrix approximation associated with the NET

        Returns
        -------
        np.ndarray (ndim = 2)
            the impedance matrix approximation
        """
        n_loc = len(self.root.loc_inds)
        loc_map = {loc_ind: map_ind for map_ind, loc_ind in enumerate(self.root.loc_inds)}
        z_mat = np.zeros((n_loc, n_loc))
        self._addNodeToImpMat(self.root, z_mat, loc_map)
        return z_mat

    def _addNodeToImpMat(self, node, z_mat, loc_map):
        inds = np.array([loc_map[loc_ind] for loc_ind in node.loc_inds])
        z_mat[np.tile(inds, len(inds)), np.repeat(inds, len(inds))] += node.z_bar
        for cnode in node.child_nodes:
            self._addNodeToImpMat(cnode, z_mat, loc_map)

    def getCompartmentalization(self, Iz, returntype='node index'):
        """
        Returns a compartmentalization for the NET tree where each pair of
        compartments is separated by an Iz of at least ``Iz``. The
        compartmentalization is coded as a list of list, each sublist representing
        a the nodes closest to the root associated with the compartment.

        Parameters
        ----------
        Iz : float
            the minimum Iz separating the compartments
        returntype: str ('node index', 'node')
            either returns the node indices or the node objects

        Returns
        -------
        list of lists
            the compartments
        """
        self._computeTentativeCompartments(Iz=Iz)
        # determine the nodes that contain the eventual compartments and
        # remove the rest
        net = copy.deepcopy(self)
        self._removeNonCompartments(net.leafs, net=net)
        # get the compartment nodes
        comp_nodes = self._setCompartmentsLeafbased(net.leafs, net)
        if returntype == 'node index':
            comp_inds  = []
            for node in comp_nodes:
                inds = node._comps[node._root_ind]
                comp_inds.append(inds)
            return comp_inds
        elif returntype == 'node':
            comp_nodes_ = []
            for node in comp_nodes:
                inds = node._comps[node.rootind]
                comp_nodes_.append([self[ind] for ind in inds])
            return comp_nodes_

    def _setCompartmentsLeafbased(self, leafs, net):
        comp_nodes = []
        for ii, leaf in enumerate(leafs):
            root, _, _ = net.sisterLeafs(leaf)
            new_leaf = leaf
            comp_bool = False
            while root.index in new_leaf._node_inds:
                comp_bool = True
                old_leaf = new_leaf
                new_leaf = old_leaf.parent_node
            if comp_bool:
                # mark the old_leaf as the compartment indexing node
                old_leaf._setSharedRootInd(root.index)
                comp_nodes.append(old_leaf)
        return comp_nodes

    def _removeNonCompartments(self, leafs, net=None, n_count=0):
        if net is None:
            warnings.warn('Modifying original tree')
            net = self
        # count number of leafs
        n_leaf = len(leafs)
        leaf = leafs[0]
        # shuffle list
        del leafs[0]
        leafs = leafs + [leaf]
        # leaf is not highest order
        common_root, sister_leafs, corresponding_children = net.sisterLeafs(leaf)
        if common_root.index == 0:
            pass
        if len(sister_leafs) == len(corresponding_children):
            # find the compartments with maximal size and closest to common root
            sleafs_comp = []
            sinds_comp = []
            for ii, leaf in enumerate(sister_leafs):
                newleaf = leaf
                comp_bool = False
                while common_root.index in newleaf._node_inds:
                    comp_bool = True
                    oldleaf = newleaf
                    newleaf = oldleaf.parent_node
                if comp_bool:
                    sinds_comp.append(ii)
                    sleafs_comp.append(oldleaf)
            # delete the leafs that are not in compartments
            if len(sleafs_comp) <= 1 and not net.isRoot(common_root):
                # if at most one is compartment, we retain only the largest one
                ind = np.argmax([self.calcTotalImpedance(node) \
                                 for node in sister_leafs])
                newleaf = sister_leafs[ind]
                for ii, cnode in enumerate(corresponding_children):
                    if ii != ind:
                        net.softRemoveNode(cnode)
                        leafs.remove(sister_leafs[ii])
            else:
                # if more can be compartments, we retain all those
                for ii, cnode in enumerate(corresponding_children):
                    if not ii in sinds_comp:
                        net.softRemoveNode(cnode)
                        leafs.remove(sister_leafs[ii])
            if n_leaf != len(leafs) and len(leafs) > 0:
                self._removeNonCompartments(leafs, net=net, n_count=0)
            elif n_count < len(leafs):
                self._removeNonCompartments(leafs, net=net, n_count=n_count+1)
        elif n_count < len(leafs) and len(leafs) > 0:
            self._removeNonCompartments(leafs, net=net, n_count=n_count+1)

    def _computeTentativeCompartments(self, Iz=5.):
        # set the prerequisite impedances
        self._setCompartmentInfo(Iz=Iz)
        # set the tentative compartments
        for node in self:
            self._setCompartmentsRelative(node)

    def _setCompartmentInfo(self, Iz=5., node=None, z_p=0.,
                        node_list=[], z_root_list=[], z_comp_list=[]):
        if node != None:
            # list of dependent impedances
            try:
                z_root_list.append(z_root_list[-1] + z_p )
            except IndexError:
                z_root_list.append(z_p)
            # list of independent impedances
            z_comp_list.append(0.)
            z_comp_list = [node.z_bar + z_c for z_c in z_comp_list]
            # list or nodes
            node_list.append(node.parent_node)
            # store the compartment information
            node._setCompartmentData(node_list, z_root_list, z_comp_list, Iz=Iz)
        else:
            node = self.root
            # compute node impedance
            self.root._setCompartmentData([], [], [], Iz=0.)
        # recurse to child nodes
        for cnode in node.child_nodes:
            self._setCompartmentInfo(Iz=Iz, node=cnode, z_p=node.z_bar,
                                node_list=copy.copy(node_list),
                                z_root_list=copy.copy(z_root_list),
                                z_comp_list=copy.copy(z_comp_list))

    def _setCompartmentsRelative(self, node):
        z_target = node._z_comp
        node_comps = []
        for z_t in z_target:
            comp = [node.index]
            node_comps.append(comp)
        node._setTentativeCompartments(node_comps)

    def computeCondRescale(self, gs):
        assert len(gs) == len(self.root.loc_inds)
        # array for storing shunt factors
        sfs = np.ones_like(gs)
        # counter for recursion algorithm
        for node in self:
            node.counter = 0
        # recursive algorithm to compute shunt factors
        self._sweep(self.leafs[0], self.leafs[1:], sfs, gs)
        # clean
        for node in self:
            node.counter = 0

        return sfs

    def _sweep(self, node, leafs, sfs, gs):
        node.counter += 1
        if node.counter >= len(node.child_nodes):
            if not self.isRoot(node):
                # compute the rescaled shunt factors
                denom = 1. + node.z_bar * np.sum(sfs[node.loc_inds] * gs[node.loc_inds])
                sfs[node.loc_inds] = sfs[node.loc_inds] / denom
                # further recursion
                self._sweep(node.parent_node, leafs, sfs, gs)
        else:
            self._sweep(leafs[0], leafs[1:], sfs, gs)

    def improveInputImpedance(self, z_mat):
        nmaxind = np.max([n.index for n in self])
        for node in self.getNodes():
            if len(node.loc_inds) == 1:
                ind = node.loc_inds[0]
                # recompute the kernel of this single loc layer
                if node.parent_node is not None:
                    p_k = self.calcTotalKernel(node.parent_node)
                else:
                    p_k = Kernel((node.z_kernel.a, np.zeros_like(node.z_kernel.a)))
                f_z = (z_mat[ind,ind] - p_k.k_bar) / node.z_bar
                node.z_kernel.c *= f_z
            elif len(node.newloc_inds) > 0:
                z_k_approx = self.calcTotalKernel(node)
                # add new input nodes for the nodes that don't have one
                tbr_inds = []
                for ind in node.newloc_inds:
                    nmaxind += 1
                    f_z = (z_mat[ind,ind] - z_k_approx.k_bar)
                    if np.abs(f_z) > 1e-7:
                        f_z /= node.z_bar
                        z_k_real = Kernel(dict(a=node.z_kernel.a, c=node.z_kernel.c*f_z))
                        # add node
                        newnode = NETNode(nmaxind, [ind], z_kernel=z_k_real)
                        newnode.newloc_inds = [ind]
                        self.addNodeWithParent(newnode, node)
                        tbr_inds.append(ind)
                for ind in tbr_inds: node.newloc_inds.remove(ind)
                # empty the new indices
                node.newloc_inds = []

        self.setNewLocInds()


    def plotDendrogram(self, ax,
                        plotargs={}, labelargs={}, textargs={},
                        incolors={},
                        inlabels={}, nodelabels={},
                        cs_comp={}, cmap=None,
                        z_max=None, add_scalebar=True):
        """
        Generate a dendrogram of the NET

        Parameters
        ----------
            ax: :class:`matplotlib.axes`
                the axes object in which the plot will be made
            plotargs : dict (string : value)
                keyword args for the matplotlib plot function, specifies the
                line properties of the dendrogram
            labelargs : dict (string : value)
                keyword args for the matplotlib plot function, specifies the
                marker properties for the node points. Or dict with keys node
                indices, and with values dicts with keyword args for the
                matplotlib function that specify the marker properties for
                specific node points. The entry under key -1 specifies the
                properties for all nodes not explicitly in the keys.
            textargs : dict (string : value)
                keyword args for matplotlib textproperties
            incolors : dict (int : string)
                dict with locinds as keys and colors as values
            inlabels : dict (int : string)
                dict with locinds as keys and label strings as values
            nodelabels: dict (int: string) or None
                labels of the nodes. If None, nodes are named by default
                according to their location indices. If empty dict, no labels
                are added.
            cs_comp : dict (int : float)
                dict with node inds as keys and compartment colors as values
            z_max: float or None
                specifies the y-scale. If None, the scale is computed from
                ``self``
            add_scalebar: bool
                whether or not to add a scale bar
        """
        if cs_comp:
            # compute the compartmental colormap if necessary
            arr = np.array([list(cs_comp.values())])
            max_cs = np.max(arr)
            min_cs = np.min(arr)
            norm_cs = (max_cs - min_cs) * (1. + 1./100.)
            for key, val in cs_comp.items():
                cs_comp[key] = (cs_comp[key] - min_cs) / norm_cs
            if cmap is None: cmap = pl.get_cmap('jet')
            cs_comp['cm'] = cmap
            Z = [[0,0],[0,0]]
            levels = np.linspace(min_cs, max_cs, 100)
            CS3 = pl.contourf(Z, levels, cmap=cmap)
        # get the number of leafs to determine the dendrogram spacing
        rnode    = self.root
        n_branch  = self.degreeOfNode(rnode)
        l_spacing = np.linspace(0., 1., n_branch+1)
        # determine input inpedances to fix the y scale
        if z_max == None:
            z_dict = {}
            for node in self.nodes:
                for ind in node.loc_inds:
                    try:
                        z_dict[ind] += node.z_bar
                    except KeyError:
                        z_dict[ind] = node.z_bar
            z_max = max(z_dict.values())
        # plot the dendrogram
        self._expandDendrogram(rnode, 0.5, 0.,
                    l_spacing, z_max, ax,
                    plotargs=plotargs, labelargs=labelargs, textargs=textargs,
                    incolors=incolors,
                    inlabels=inlabels, nodelabels=nodelabels,
                    cs_comp=cs_comp)
        # limits
        ax.set_ylim((0.0, 1.2*z_max))
        ax.set_xlim((0.,1.))
        # scalebar
        if add_scalebar:
            sblength = np.around(z_max // 5, -2)
            if sblength < .1: sblength += np.around(z_max % 5, -1)
            if sblength < .1: sblength += np.around(z_max // 5, 0)
            sbwidth = plotargs['lw']*3 if 'lw' in plotargs else 3.
            sbtsize = textargs['size'] if 'size' in textargs else 'small'
            ax.plot([0.,0.], [0., sblength], 'k-', lw=sbwidth)
            ax.annotate(r'%.0f M$\Omega$'%sblength,
                            xy=(0., sblength/2.), xytext=(-0.04, sblength/2.),
                            size=sbtsize, rotation=90, ha='center', va='center')

        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)
        ax.axison = False

        return z_max

    def _expandDendrogram(self, node, x0, y0,
                                        l_spacing, z_max, ax,
                                        plotargs={}, labelargs={}, textargs={},
                                        incolors={},
                                        inlabels={}, nodelabels={},
                                        cs_comp={}):
        # check if part of compartment
        if cs_comp:
            if node.index in list(cs_comp.keys()):
                plotargs = copy.deepcopy(plotargs)
                plotargs['color'] = cs_comp['cm'](cs_comp[node.index])
        # impedance of layer
        ynew = y0 + node.z_bar
        # plot vertical connection line
        ax.vlines(x0, y0, ynew, **plotargs)
        # get the child nodes for recursion
        l0 = 0
        for i, cnode in enumerate(node.child_nodes):
            # attribute space on xaxis
            deg = self.degreeOfNode(cnode)
            l1 = l0 + deg
            # new quantities
            xnew = (l_spacing[l0] + l_spacing[l1]) / 2.
            # horizontal connection line limits
            if i == 0:
                xnew0 = xnew
            if i == len(node.child_nodes)-1:
                xnew1 = xnew
            # recursion
            self._expandDendrogram(cnode, xnew, ynew,
                    l_spacing[l0:l1+1], z_max, ax,
                    plotargs=plotargs, labelargs=labelargs, textargs=textargs,
                    incolors=incolors,
                    inlabels=inlabels, nodelabels=nodelabels,
                    cs_comp=cs_comp)
            # next index
            l0 = l1
        # plot horizontal connection line
        if l0 > 0:
            ax.hlines(ynew, xnew0, xnew1, **plotargs)
        # add label and maybe text annotation to node
        if node.index in labelargs:
            ax.plot([x0], [ynew], **labelargs[node.index])
        elif -1 in labelargs:
            ax.plot([x0], [ynew], **labelargs[-1])
        else:
            try:
                ax.plot([x0], [ynew], **labelargs)
            except TypeError as e:
                pass
        if textargs:
            if nodelabels != None:
                if node.index in nodelabels:
                    if labelargs == {}:
                        ax.plot([x0], [ynew], **nodelabels[node.index][1])
                        ax.annotate(nodelabels[node.index][0],
                                    xy=(x0, ynew), xytext=(x0+0.04, ynew+z_max*0.04),
                                    bbox=dict(boxstyle='round', ec=(1., 0.5, 0.5), fc=(1., 0.8, 0.8)),
                                    **textargs)
                    else:
                        ax.annotate(nodelabels[node.index],
                                    xy=(x0, ynew), xytext=(x0+0.04, ynew+z_max*0.04),
                                    bbox=dict(boxstyle='round', ec=(1., 0.5, 0.5), fc=(1., 0.8, 0.8)),
                                    **textargs)
            else:
                ax.annotate(r'$N='+''.join([str(ind) for ind in node.loc_inds])+'$',
                                 xy=(x0, ynew), xytext=(x0+0.04, ynew+z_max*0.04),
                                 bbox=dict(boxstyle='round', ec=(1., 0.5, 0.5), fc=(1., 0.8, 0.8)),
                                 **textargs)
            # add input label
            if self.isLeaf(node):
                if inlabels != None:
                    lwidth = plotargs['lw'] if 'lw' in plotargs else 1.
                    ax.vlines(x0, ynew+z_max*0.04, z_max*1.1, lw=lwidth, linestyle=':', color='k')
                    if node.loc_inds[0] in incolors:
                        bboxdict = dict(boxstyle='round', ec=incolors[node.loc_inds[0]], fc=incolors[node.loc_inds[0]], alpha=0.5)
                    else:
                        bboxdict = dict(boxstyle='round', ec=(0.5, 0.5, 1.), fc=(0.8, 0.8, 1.))
                    if node.loc_inds[0] in inlabels:
                        textstr = inlabels[node.loc_inds[0]]
                    else:
                        textstr = r'$'+str(node.loc_inds[0])+'$'
                    ax.annotate(textstr,
                                        xy=(x0, z_max*1.1), xytext=(x0, z_max*1.14), ha='center',
                                        bbox=bboxdict,
                                        **textargs)



"""
File contains:

    - :class:`NETSim`

Author: W. Wybo
"""


cimport numpy as np
import numpy as np

from libcpp.string cimport string
from libc.stdint cimport int16_t, int32_t, int64_t
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.deque cimport deque
from libcpp.map cimport map
from libcpp.pair cimport pair

import copy
import time
import warnings

from neat.channels.channelcollection import channelcollection


def c2r(arr_c):
    return np.concatenate((arr_c.real[:, np.newaxis], arr_c.imag[:, np.newaxis]), 1)

cdef extern from "NETC.h":
    cdef cppclass NETSimulator:
        NETSimulator(int n_loc, double *v_eq)

        # initialize from python
        void initFromPython(double dt, double integ_mode, bool print_tree)
        void addNodeFromPython(int node_index, int parent_index,
                               int64_t *child_indices, int n_children,
                               int64_t *loc_indices, int n_locinds,
                               int64_t *newloc_indices, int n_newlocinds,
                               double *alphas, double *gammas, int n_exp)
        void addLinTermFromPython(int loc_index,
                                double *alphas, double *gammas, int n_exp)
        void addIonChannelFromPython(string channel_name, int loc_ind,
                                     double g_bar, double e_rev,
                                     bool instantaneous, double *vs, int v_size);
        void addSynapseFromType(int loc_ind, int syn_type)
        void addSynapseFromParams(int loc_ind, double e_r,
                                double *params, int p_size)
        void reset()
        # other structure functions
        void removeSynapseFromIndex(int loc_ind, int syn_ind)
        # getter functions for voltage
        void addVLocToArr(double *v_arr, int v_size)
        void addVNodeToArr(double *v_arr, int v_size)
        double getVSingleLoc(int loc_index)
        double getVSingleNode(int node_index)
        double getGSingleSyn(int loc_index, int syn_index)
        double getSurfaceSingleSyn(int loc_index, int syn_index)
        void setVNodeFromVLoc(double *v_arr, int v_size)
        void setVNodeFromVNode(double *v_arr, int v_size)
        # integration functions
        void solveMatrix()
        void constructMatrix(double dt, double* mat, double* arr, int n_node);
        void setInputsToZero()
        void constructInputSyn1Loc(int loc_ind, double v_m,
                                double *g_syn, int g_size)
        void constructInputChan1Loc(int loc_ind, double v_m)
        void advance(double dt)
        void feedSpike(int loc_ind, int syn_ind, double g_max, int n_spike)


cdef class NETSim:
    cdef NETSimulator *net_ptr
    # numbers
    cdef public int n_loc
    cdef public int n_node
    cdef public np.ndarray n_syn
    # equilibirum potentials
    cdef public np.ndarray v_eq
    # flag indicating Newton solver (0) or integration with full kernels (1)
    cdef int mode
    # store the synapses, conductances and spikes
    cdef public list syn_map_py
    cdef public list spike_times_py

    def __cinit__(self, net, lin_terms=None, v_eq=-75., print_tree=False):
        self.n_loc = len(net.root.loc_inds)
        # equibrium potentials
        if isinstance(v_eq, float):
            self.v_eq = v_eq * np.ones(self.n_loc)
        elif len(v_eq) == self.n_loc:
            self.v_eq = np.array(v_eq)
        else:
            raise IOError("Invalid value for input argument [v_eq] should be float " + \
                          "or iterable of floats")
        cdef np.ndarray[np.double_t, ndim=1] v_eq_arr = self.v_eq
        # temporary python objects
        self.syn_map_py = []
        self.spike_times_py = []
        # create the cnet
        self.n_node = len(net)
        for node in net:
            if node.parent_node is not None:
                pnode_ind = node.parent_node.index
            else:
                pnode_ind = -1
                # initialize the C++ object
                self.net_ptr = new NETSimulator(self.n_loc, &v_eq_arr[0])
                self.n_syn = np.zeros(self.n_loc, dtype=int)
            cnode_inds = np.array([cnode.index for cnode in node.child_nodes], dtype=int)
            if len(cnode_inds) == 0: cnode_inds = np.array([-1])
            locinds = np.array(list(node.loc_inds)) \
                                if len(node.loc_inds) > 0 else np.array([-1])
            newlocinds = np.array(list(node.newloc_inds)) \
                                if len(node.newloc_inds) > 0 else np.array([-1])
            (alphas, gammas) = (-node.z_kernel.a, node.z_kernel.c)
            alphas = alphas.astype(complex); gammas = gammas.astype(complex)
            self._addNode(node.index, pnode_ind,
                            cnode_inds,
                            locinds,
                            newlocinds,
                            c2r(alphas), c2r(gammas))
        # add the linear terms
        if lin_terms is not None:
            if isinstance(lin_terms, dict):
                for ind, lin_term in lin_terms.iteritems():
                    alphas, gammas = -lin_term.a, lin_term.c
                    alphas, gammas = alphas.astype(complex), gammas.astype(complex)
                    self._addLinTerm(ind, c2r(alphas), c2r(gammas))
            else:
                raise IOError('`lin_terms` should be dict')
        self.initialize(mode=0, print_tree=print_tree)

    def initialize(self, dt=.1, mode=1, print_tree=False):
        self.mode = mode
        self.net_ptr.initFromPython(dt, mode, print_tree)

    def _addNode(self, node_index, parent_index,
                       np.ndarray[np.int64_t, ndim=1] child_indices,
                       np.ndarray[np.int64_t, ndim=1] loc_indices,
                       np.ndarray[np.int64_t, ndim=1] newloc_indices,
                       np.ndarray[np.double_t, ndim=2] alphas,
                       np.ndarray[np.double_t, ndim=2] gammas):
        self.net_ptr.addNodeFromPython(node_index, parent_index,
                                   &child_indices[0], child_indices.shape[0],
                                   &loc_indices[0], loc_indices.shape[0],
                                   &newloc_indices[0], newloc_indices.shape[0],
                                   &alphas[0,0], &gammas[0,0], alphas.shape[0])

    def _addLinTerm(self, loc_index,
                       np.ndarray[np.double_t, ndim=2] alphas,
                       np.ndarray[np.double_t, ndim=2] gammas):
        self.net_ptr.addLinTermFromPython(loc_index,
                                    &alphas[0,0], &gammas[0,0], alphas.shape[0])

    def setVNodeFromVNode(self, v_node):
        """
        Set the node voltage

        Parameters
        ----------
        v_loc : ndarray (`ndim` = 1, `size` = ``n_node``)
            node voltage
        """
        if v_node.shape[0] != self.n_node:
            raise ValueError('Input has incorrect size')
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = v_node
        self.net_ptr.setVNodeFromVNode(&vc_arr[0], vc_arr.shape[0])

    def setVNodeFromVLoc(self, v_loc):
        """
        Set the node voltage starting from the location voltage array. This
        transformation is not unique. The implementation chosen here is so that
        nodes which integrate new locations receive all of that locations
        voltage and the underlying nodes none

        Parameters
        ----------
        v_loc : ndarray (`ndim` = 1, `size` = ``n_loc``)
            location voltage
        """
        if v_loc.shape[0] != self.n_loc:
            raise ValueError('Input has incorrect size')
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = v_loc
        self.net_ptr.setVNodeFromVLoc(&vc_arr[0], vc_arr.shape[0])

    def getVLoc(self):
        """
        Returns the location voltages

        Returns
        -------
        numpy.ndarray (`ndim` = 1, `size` = ``n_loc``)
        """
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = np.zeros((self.n_loc,), dtype=float)
        self.net_ptr.addVLocToArr(&vc_arr[0], vc_arr.shape[0])
        return vc_arr

    def addVLocToArr(self, v_arr):
        """
        Add the location voltages to the input array

        Parameters
        ----------
        v_arr : ndarray (`ndim` = 1, `size` = ``n_loc``)
        """
        if v_arr.shape[0] != self.n_loc:
            raise ValueError('Input has incorrect size')
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = v_arr
        self.net_ptr.addVLocToArr(&vc_arr[0], vc_arr.shape[0])

    def getVNode(self):
        """
        Returns the NET node voltages

        Returns
        -------
        numpy.ndarray (`ndim` = 1, `size` = ``n_node``)
        """
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = np.zeros((self.n_node,), dtype=float)
        self.net_ptr.addVNodeToArr(&vc_arr[0], vc_arr.shape[0])
        return vc_arr

    def addVNodeToArr(self, v_arr):
        """
        Add the NET node voltages to the input array

        Parameters
        ----------
        v_arr : ndarray (`ndim` = 1, `size` = ``n_node``)
        """
        if v_arr.shape[0] != self.n_node:
            raise ValueError('Input array has incorrect size')
        cdef np.ndarray[np.double_t, ndim=1] vc_arr = v_arr
        self.net_ptr.addVNodeToArr(&vc_arr[0], vc_arr.shape[0])

    def addChannel(self, chan, loc_index, g_max, e_rev,
                         instantaneous=False, v_const={}):
        """
        Add an ion channel to the NET model

        Parameters
        ----------
        chan: `neat.channels.IonChannel
            name of the ion channel
        loc_index: int
            index of the location
        g_max: float
            maximal conductance of the channel
        e_rev: float
            reversal potential (if not provide, default value is taken)
        instantaneous: bool (optional, default to 'False')
            whether the activation should be approximated as instantaneous if
            such a state variable is present
        v_const: dict {string: float} (optional)
            state variables for which a voltage value is provided (< 1000 mV)
            are fixed at their activation levels for this voltage value during
            the Newton iteration. If the voltage value exceeds 1000 mV or is not
            provided, state variable activations are computed based on actual
            voltage during Newton iteration
        """
        cdef string cname = chan.__class__.__name__.encode('UTF-8')
        # voltage values
        cdef np.ndarray[np.double_t, ndim=1] vs_arr = np.zeros(len(chan.statevars))
        try:
            for ii, svar in enumerate(chan.ordered_statevars):
                sv = str(svar)
                vs_arr[ii] = v_const[sv] if sv in v_const else 1e4
        except AttributeError:
            # backwards compatible with old style channels
            for ii, svar in enumerate(chan.varnames.flatten()):
                vs_arr[ii] = v_const[svar] if svar in v_const else 1e4
        # add the channel
        self.net_ptr.addIonChannelFromPython(cname, loc_index, g_max, e_rev,
                                             instantaneous, &vs_arr[0], vs_arr.shape[0])

    def addSynapse(self, loc_index, synarg, g_max=0.001, nmda_ratio=1.6):
        """
        Add a synapse to the NET model

        Parameters
        ----------
        loc_index : int
            index of the synapse location
        synarg : float, string or dict
            for `float`, specifies the synaptic reversal potential (mV)
            for `string`, choos 'AMPA', 'NMDA' or 'GABA' or 'AMPA+NMDA'
            for `dict`, contains either three entries ('e_r' (reversal potential),
                'tau_r' (rise time), 'tau_d' (decay time)) or two entries ('e_r'
                (reversal potential), 'tau' (decay time))
        """
        cdef np.ndarray[np.double_t, ndim=1] tau_arr
        if loc_index < 0 or loc_index >= self.n_loc:
            raise IndexError('`loc_index` out of range')
        if isinstance(synarg, str):
            if synarg == 'AMPA+NMDA':
                # add the synapses
                self.net_ptr.addSynapseFromType(loc_index, 0)
                self.net_ptr.addSynapseFromType(loc_index, 1)
                # synapse parameters for simulation
                syn_map = {'loc_index': loc_index,
                            'syn_index_at_loc': self.n_syn[loc_index],
                            'n_syn_at_loc': 2,
                            'g_max': [g_max, nmda_ratio*g_max]}
                self.n_syn[loc_index] += 2
            else:
                if synarg == 'AMPA':
                    synarg = 0
                elif synarg == 'NMDA':
                    synarg = 1
                elif synarg == 'GABA':
                    synarg = 2
                else:
                    raise ValueError('``synarg`` should be \'AMPA\', ' \
                                        '\'NMDA\' or \'GABA\' or \'AMPA+NMDA\'')
                # add the synapse
                self.net_ptr.addSynapseFromType(loc_index, synarg)
                # synapse parameters for simulation
                syn_map = {'loc_index': loc_index,
                            'syn_index_at_loc': self.n_syn[loc_index],
                            'n_syn_at_loc': 1,
                            'g_max': [g_max]}
                self.n_syn[loc_index] += 1
        elif isinstance(synarg, float):
            tau_arr = np.array([.2,3.])
            # add the synapse
            self.net_ptr.addSynapseFromParams(loc_index, synarg,
                                              &tau_arr[0], tau_arr.shape[0])
            # synapse parameters for simulation
            syn_map = {'loc_index': loc_index,
                        'syn_index_at_loc': self.n_syn[loc_index],
                        'n_syn_at_loc': 1,
                        'g_max': [g_max]}
            self.n_syn[loc_index] += 1
        elif isinstance(synarg, dict):
            if 'tau_r' in synarg and 'tau_d' in synarg:
                tau_arr = np.array([synarg['tau_r'], synarg['tau_d']])
            elif 'tau' in synarg:
                tau_arr = np.array([synarg['tau']])
            else:
                raise KeyError('No time scale keys found in `dict` ``synarg`` ' + \
                                'use \'tau\' or \'tau_r\' and \'tau_d\'')
            # add the synapse
            self.net_ptr.addSynapseFromParams(loc_index, synarg['e_r'],
                                              &tau_arr[0], tau_arr.shape[0])
            # synapse parameters for simulation
            syn_map = {'loc_index': loc_index,
                        'syn_index_at_loc': self.n_syn[loc_index],
                        'n_syn_at_loc': 1,
                        'g_max': [g_max]}
            self.n_syn[loc_index] += 1
        else:
            raise TypeError('``synarg`` should be `string` or `float` or `dict`')

        self.syn_map_py.append(syn_map)
        self.spike_times_py.append([])

    def _getSynListIndex(self, loc_index, syn_index):
        if loc_index < 0 or loc_index >= self.n_loc:
            raise IndexError('`loc_index` out of range')
        if syn_index < 0 or syn_index >= self.n_syn[loc_index]:
            raise IndexError('`syn_index` out of range')
        # find the synapse index and return it
        for ii, syn_map in enumerate(self.syn_map_py):
            if syn_map['loc_index'] == loc_index and \
               syn_map['syn_index_at_loc'] == syn_index:
                return ii
        return None

    def getCondSurfFromLoc(self, loc_index, syn_index):
        return self.net_ptr.getSurfaceSingleSyn(loc_index, syn_index)

    def getCondSurf(self, index):
        return self.net_ptr.getSurfaceSingleSyn(self.syn_map[index]['loc_index'], \
                                               self.syn_map[index]['syn_index_at_loc'])

    def removeSynapseFromLoc(self, loc_index, syn_index):
        """
        Remove synapse at location ``loc_index``, and at position ``syn_index`` in
        the synapse array of that location

        Parameters
        ----------
        loc_index : int
            index of the synapse location
        syn_index : int
            index of the synapse
        """
        index = self._getSynListIndex(loc_index, syn_index)
        self.removeSynapse(index)

    def removeSynapse(self, index):
        """
        Remove synapse at position ``index`` in synapse stack

        Parameters
        ----------
        index : int
            index of the synapse
        """
        if not isinstance(index, int):
            raise ValueError('Expected int')
        elif index < 0 or index > len(self.syn_map_py):
            raise IndexError('\'index\' out of range')
        syn_map_del = self.syn_map_py.pop(index)
        self.n_syn[syn_map_del['loc_index']] -= syn_map_del['n_syn_at_loc']
        for _ in xrange(syn_map_del['n_syn_at_loc']):
            self.net_ptr.removeSynapseFromIndex(syn_map_del['loc_index'],
                                                syn_map_del['syn_index_at_loc'])
        # reassign indices of other spike times
        del self.spike_times_py[index]

    def _constructInput(self, v_arr, g_syn):
        """
        constructs the NET solver input for the given conductance and voltage

        Parameters
        ----------
        g_syn : list of np.ndarray (`ndim` = 1)
            list contains ``self.n_loc`` arrays, each containing the
            conductances at the corresponding location
        v_arr : np.ndarray (`ndim` = 1)
            the location voltage
        """
        if v_arr.shape[0] != self.n_loc:
            raise ValueError('`v_arr` has incorrect size')
        if len(g_syn) != self.n_loc:
            raise ValueError('`g_syn` has incorrect size')
        # declare c array
        cdef np.ndarray[np.double_t, ndim=1] g_c
        # reset the arrays
        self.net_ptr.setInputsToZero()
        for ii, (v_a, g_s) in enumerate(zip(v_arr, g_syn)):
            if g_s.size != self.n_syn[ii]:
                raise ValueError('`g_syn[%d]` has incorrect size'%ii)
            if g_s.size > 0:
                g_c = g_s
                self.net_ptr.constructInputSyn1Loc(ii, v_a, &g_c[0], g_c.shape[0])
            self.net_ptr.constructInputChan1Loc(ii, v_a)

    def recastInput(self, g_syn):
        """
        executes the Newton iteration to find the steady state location voltage

        Parameters
        ----------
        g_syn : list of np.ndarray (`ndim`=1) or single iterable over floats
            if list, contains ``self.n_loc`` arrays (`ndim`=1), each containing
            the conductances of synapses at the corresponding location
            if single iterable, contains conductances of each synapse on the
            synapse stack

        Returns
        -------
        list of np.ndarray (`ndim`=1)

        """
        # recast input in other form
        if not hasattr(g_syn[0], '__iter__'):
            g_syn_ = g_syn
            g_syn = [[] for _ in xrange(self.n_loc)]
            for ii, syn_map in enumerate(self.syn_map_py):
                g_syn[syn_map['loc_index']].append(g_syn_[ii])
                # automatically assumes second synapse (if present) is NMDA
                # synapse, multiplies maximal conductance with NMDA ratio
                surf_ampa = self.net_ptr.getSurfaceSingleSyn(\
                            syn_map['loc_index'], syn_map['syn_index_at_loc'])
                for jj, g_s in enumerate(syn_map['g_max'][1:]):
                    surf_nmda = self.net_ptr.getSurfaceSingleSyn(\
                            syn_map['loc_index'], syn_map['syn_index_at_loc']+jj+1)
                    g_syn[syn_map['loc_index']].append( \
                                g_syn_[ii] * g_s / syn_map['g_max'][0] * \
                                surf_nmda / surf_ampa)
            g_syn = [np.array(g_s) for g_s in g_syn]
            return g_syn
        else:
            return g_syn

    def getMatAndVec(self, dt=0.1):
        # construct the data arrays
        cdef np.ndarray[np.double_t, ndim=2] mat = np.zeros((self.n_node, self.n_node))
        cdef np.ndarray[np.double_t, ndim=1] vec = np.zeros(self.n_node)
        # fill matrix and vector
        self.net_ptr.constructMatrix(dt, &mat[0,0], &vec[0], self.n_node)
        return mat, vec

    def invertMatrix(self):
        self.net_ptr.solveMatrix()

    def solveNewton(self, g_syn,
                           v_eps=.1, n_max=100,
                           v_0=None, v_alt=None,
                           n_iter=0):
        """
        executes the Newton iteration to find the steady state location voltage

        Parameters
        ----------
        g_syn : list of np.ndarray (`ndim` = 1) or single iterable over floats
            if list, contains ``self.n_loc`` arrays (`ndim` = 1), each containing
            the conductances of synapses at the corresponding location
            if single iterable, contains conductances of synapses synapses
            on the synapse stack
        v_eps : float (mV), optional
            the tolerance for convergence
        n_max : int, optional
            maximum iteration number
        v_0 : np.ndarray, optional
            the initial location voltage. De
        v_alt : np.ndarray, optional
            initial location voltage when iteration starting from ``v_0`` fails
            to converge
        n_iter : int
            iteration number, for recursion, do not touch

        Returns
        -------
        np.ndarray (`ndim` = 1)
            the location voltage
        """
        # check if NET simulator is in correct integration mode
        if self.mode != 0:
            self.initialize(mode=0)
        # location voltage
        if v_0 is None:
            v_0 = self.v_eq
            v_alt = np.zeros(self.n_loc)
        if n_iter == 0:
            self.setVNodeFromVLoc(v_0)
            # recast input in other form
            g_syn = self.recastInput(g_syn)
        # construct the inputs
        self._constructInput(v_0, g_syn)
        # compute solution
        self.net_ptr.solveMatrix()
        v_new = self.getVLoc()
        # check for instability
        if np.max(np.abs(v_new)) > 1000.:
            # v_0 failed, so try v_alt
            if v_alt is not None:
                self.setVNodeFromVLoc(v_alt)
                v_new = self.solveNewton(g_syn, v_eps=v_eps,
                                            n_iter=0, n_max=n_max,
                                            v_0=v_alt, v_alt=None)
            else:
                warnings.warn('Newton solver failed to converge')
                return np.array([np.nan for _ in range(self.n_loc)])
        elif np.linalg.norm(v_new - v_0) > v_eps and n_iter < n_max:
            # no convergence yet, move on to next step
            v_new = self.solveNewton(g_syn, v_eps=v_eps,
                                        n_iter=n_iter+1, n_max=n_max,
                                        v_0=v_new, v_alt=v_alt)
        return v_new

    def setSpikeTimesFromLoc(self, loc_index, syn_index, spike_times):
        """
        Set spike times for a synapse indexed by it's location ``loc_index`` and
        position ``syn_index`` at that location

        Parameters
        ----------
        loc_index : int
            index of the synapse location
        syn_index : int
            index of the synapse
        spike_times : iterable
            iterable collection of spike times. Needs to be ordered
        """
        index = self._getSynListIndex(loc_index, syn_index)
        self.setSpikeTimes(index, spike_times)

    def setSpikeTimes(self, index, spike_times):
        """
        Set spike times for a synapse indexed by it's position in the synapse
        stack

        Parameters
        ----------
        index : int
            position of synapse in synapse stack
        spike_times : iterable
            iterable collection of spike times. Needs to be ordered. Can be
            empty.
        """
        if not isinstance(index, int):
            raise ValueError('Expected int')
        elif index < 0 or index > len(self.syn_map_py):
            raise IndexError('\'index\' out of range')
        self.spike_times_py[index] = spike_times

    def runSim(self, double tmax, double dt, int step_skip=1,
                    bool rec_v_node=False, list rec_g_syn_inds=[], bool pprint=False):
        """
        Run the simulation using the CNET.

        Parameters
        ----------
        tmax : float
            The simulation time
        dt : float
            The simulation time step
        step_skip : int
            The number of time steps to skip between storing the variables
        rec_v_node : bool
            whether to recorde the node voltage
        rec_g_syn_inds : list of inds
            record the conductance of the synapses indexed by the entries in
            this list
        pprint : bool
            whether or not to print information

        Returns
        -------
        dict
            The simulation results. Contains the keys:
            't' : np.ndarray (ndim = 1)
                contain the times at which the variables have been stored
            'v_loc' : np.ndarray (ndim = 2)
                the location voltages during the simulation, dim 1 are the
                locations, dim 2 the simulation time
            'v_node' : np.ndarray (ndim = 2, only if ``rec_v_node`` is ``True``)
                the node voltages during the simulation, dim 1 are the nodes,
                dim 2 the simulation time
            'g_syn' : np.ndarray (ndim = 2, only if ``rec_g_syn_inds`` is not empty)
                the synaptic conductances, dim 1 corresponds to indices in
                ``rec_g_syn_inds`` and dim 2 is the simulation time
        """
        if dt < 0: raise ValueError('time step has to be positive')
        if tmax < dt: raise ValueError('maximum time be larger than time step')
        cdef int k_max = int(tmax / dt)
        cdef int k_store = k_max // step_skip
        if step_skip < 0 or step_skip > k_max:
            raise ValueError('\'step_skip*dt\' has to be larger than \'tmax\'')
        # generic loop indices
        cdef int ii, jj, kk, ll, mm
        # initialize
        self.initialize(dt=dt, mode=1)
        # reset all state variables in the model
        self.net_ptr.reset()
        # convert syn map in efficient c vectors
        cdef int n_syn = len(self.syn_map_py)
        cdef vector[vector[double]] g_max = \
                            [sm['g_max'] for sm in self.syn_map_py]
        cdef vector[int] n_syn_at_loc = \
                            [sm['n_syn_at_loc'] for sm in self.syn_map_py]
        cdef vector[int] syn_index_at_loc = \
                            [sm['syn_index_at_loc'] for sm in self.syn_map_py]
        cdef vector[int] loc_index = \
                            [sm['loc_index'] for sm in self.syn_map_py]
        # convert spike times to format stored in vector of c deques
        cdef vector[deque[double]] spike_times
        cdef deque[double] spike_deque
        cdef double spike_time
        for ii, spike_times_aux in enumerate(self.spike_times_py):
            spike_times.push_back(spike_deque)
            for spike_time in spike_times_aux:
                spike_times[ii].push_back(spike_time)
            spike_deque.clear()
        # create arrays for storage
        cdef np.ndarray[np.double_t, ndim=1] t_sim = \
                    np.arange(0., (float(k_store*step_skip)-0.5)*dt, dt*step_skip)
        cdef np.ndarray[np.double_t, ndim=2] v_loc = \
                    np.zeros((self.n_loc, k_store), dtype=float)
        cdef np.ndarray[np.double_t, ndim=2] v_node
        if rec_v_node: v_node = np.zeros((self.n_node, k_store), dtype=float)
        cdef dict g_syn = {ii: [np.zeros(k_store) for _ in xrange(n_syn_at_loc[ii])] \
                                                  for ii in rec_g_syn_inds}

        if pprint:
            print('\n>>> Integrating NET model for ' + str(tmax) + ' ms. <<<')
            start = time.process_time()

        # cython loop
        cdef int ss = 0 # spike counter
        cdef double tt = 0.0 # time
        mm = 0 # store counter
        store_ind = step_skip - 1 # index when to store
        for kk in xrange(k_max):
            tt = float(kk) * dt
            # loop over all the synapses to feed input spikes
            for jj in xrange(n_syn):
                ss = 0
                while not spike_times[jj].empty() and spike_times[jj][0] < tt:
                    ss += 1
                    spike_times[jj].pop_front()
                if ss != 0:
                    for ll in xrange(n_syn_at_loc[jj]):
                        self.net_ptr.feedSpike(loc_index[jj], syn_index_at_loc[jj]+ll,
                                                g_max[jj][ll], ss)
            # advance the model one time step
            self.net_ptr.advance(dt)
            # store the variables
            if kk % step_skip == store_ind:
                for ll in xrange(self.n_loc):
                    v_loc[ll,mm] = self.net_ptr.getVSingleLoc(ll)
                if rec_v_node:
                    for ll in xrange(self.n_node):
                        v_node[ll,mm] = self.net_ptr.getVSingleNode(ll)
                for ii, g_store in g_syn.iteritems():
                    for ll in xrange(n_syn_at_loc[ii]):
                        g_store[ll][mm] = \
                                self.net_ptr.getGSingleSyn(loc_index[ii],
                                                           syn_index_at_loc[ii]+ll)
                mm += 1

        if pprint:
            stop = time.process_time()
            print('>>> Elapsed time: ' + str(stop-start) + ' seconds. <<<\n')

        #return the result
        resdict = {'t': t_sim, 'v_loc': v_loc}
        if rec_v_node:
            resdict['v_node'] = v_node
        if len(rec_g_syn_inds) > 0:
            resdict['g_syn'] = g_syn
        return resdict